import itertools
from uuid import uuid4
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

from glypy.composition import Composition
from glypy.utils import uid, basestring, make_struct
from .base import SaccharideBase, SubstituentBase
from .constants import UnknownPosition, LinkageType


default_parent_loss = Composition({"O": 1, "H": 1})
default_child_loss = Composition(H=1)


linkage_configuration = make_struct("linkage_configuration", ("parent", "child", "parent_position", "child_position"))


class Link(object):
    '''
    Represents the linkage between two molecules, described as an edge in a graph with optional
    directedness semantics.

    Attributes
    ----------
    parent: :class:`~.Monosaccharide` or :class:`~.Substituent`
    child: :class:`~.Monosaccharide` or :class:`~.Substituent`
    parent_position: int
        Position of link on :attr:`parent`
    child_position: int
        position of link on :attr:`child`
    parent_loss: |Composition|
        Elemental composition lost from the :attr:`parent` when forming this bond
    child_loss: |Composition|
        Elemental composition lost from the :attr:`child` when forming this bond
    _attached: :class:`bool`
        An internal flag that keeps track of whether or not a link is attached to
        its termini. Should not be read or manipulated directly. Instead see :meth:`is_attached`.
    '''

    __slots__ = (
        "parent", "child", "parent_position", "child_position",
        "parent_loss", "child_loss", "id",
        "label", "_attached",
        "parent_linkage_type",
        "child_linkage_type"
    )

    def __init__(self, parent, child, parent_position=UnknownPosition, child_position=UnknownPosition,
                 parent_loss=None, child_loss=None, id=None, attach=True, parent_linkage_type=None,
                 child_linkage_type=None):
        '''
        Defines a bond between `parent` and `child` between the molecule positions specified.
        The bond may represent a partial loss of elemental composition from the parent and/or child
        molecules.

        Instantiating the |Link| object will automatically attach it to its endpoints, mutating them
        unless `attach=False`. If not attached on instantiation, the bond can be realized by calling
        :meth:`Link.apply` at a later time.

        Parameters
        ----------
        parent: :class:`~.Monosaccharide` or :class:`~.Substituent`
        child: :class:`~.Monosaccharide` or :class:`~.Substituent`
        parent_position: int
            The position on the parent to attach to Defaults to :const:`~.UnknownPosition`
        child_position: int
            The position on the child to attach to. Defaults to :const:`~.UnknownPosition`
        parent_loss: :class:`Composition` or str
            The elemental composition deducted from the parent when the bond is applied
        child_loss: :class:`Composition` or str
            The elemental composition deducted from the child when the bond is applied
        id: int
            A locally unique identifier within a graph. If |None|, :func:`~.glypy.utils.uid` is used to generate one. Defaults to |None|
        attach: bool
            Whether to immediately attach the |Link| object to the `parent` and `child` molecules on instantiation
            by using :meth:`Link.apply`
        '''

        if id is None:
            id = uid()

        if parent_loss is None:
            parent_loss = default_parent_loss
        elif isinstance(parent_loss, basestring):
            parent_loss = Composition(formula=parent_loss)
        if child_loss is None:
            child_loss = default_child_loss
        elif isinstance(child_loss, basestring):
            child_loss = Composition(formula=child_loss)

        self.parent = parent
        self.child = child
        self.parent_position = parent_position
        self.child_position = child_position
        self.parent_loss = parent_loss
        self.child_loss = child_loss
        self.parent_linkage_type = parent_linkage_type or LinkageType.unknown
        self.child_linkage_type = child_linkage_type or LinkageType.unknown
        self.id = id
        self.label = None
        self._attached = False
        if attach:
            self.apply()

    def has_ambiguous_linkage(self):
        """Returns whether the link position at either terminal is ambiguous, but not unknown.

        Returns
        -------
        bool
        """
        return False

    def has_ambiguous_termini(self):
        """Returns whether the link terminal is ambiguous

        Returns
        -------
        bool
        """
        return False

    def is_ambiguous(self):
        """Returns if the link position or link terminal at either end is ambiguous,
        but not unknown.

        Returns
        -------
        bool

        See Also
        --------
        :meth:`has_ambiguous_linkage`
        :meth:`has_ambiguous_termini`
        """
        return self.has_ambiguous_linkage() or self.has_ambiguous_termini()

    def apply(self):
        '''
        Actually alter :attr:`parent` and :attr:`child`'s state. Deduct their respective losses from their
        :attr:`composition` and insert a reference to :obj:`self` into their :attr:`links` or :attr:`substituent_links`
        as appropriate.

        This performs no position availability validation. If there is a possibility that either the parent or child
        position may be occupied, they should be tested using :meth:`~.Monosaccharide.is_occupied` prior to calling
        this method.

        Sets :attr:`_attached` to |True|

        See also
        --------
        :meth:`Link.break_link`
        :meth:`Link.refund`
        :meth:`Monosaccharide.is_occupied`
        '''
        # assert not self.is_attached(), ("Cannot apply an already attached link")
        self.parent.composition -= (self.parent_loss)

        self.child.composition -= (self.child_loss)
        if self.is_substituent_link():
            self.parent.substituent_links[self.parent_position] = self
        else:
            self.parent.links[self.parent_position] = self
        self.child.links[self.child_position] = self
        self.parent._degree += 1
        self.child._degree += 1
        self._attached = True

    def to(self, mol):
        '''
        Traverse the link from `mol` to its adjacent node. If `mol` is not `self.parent` or
        `self.child`, instead raises an error. :meth:`__getitem__` is an alias of `to`

        Parameters
        ----------
        mol: :class:`~.Monosaccharide` or :class:`~.Substituent`

        Returns
        -------
        :class:`~.Monosaccharide` or :class:`~.Substituent`

        Raises
        ------
        KeyError
            If `mol` is not the parent or child of this |Link|
        '''
        if mol is (self.parent):
            return self.child
        elif mol is (self.child):
            return self.parent
        else:
            raise KeyError("Could not find connection for {0}".format(mol))

    #: Alias for :meth:`to`
    __getitem__ = to

    def __iter__(self):
        '''
        Yields the `parent` node, then the `child` node
        '''
        yield self.parent
        yield self.child

    def is_substituent_link(self):
        '''
        If :attr:`child` is a |Substituent| and :attr:`parent` is a |Monosaccharide|, then `self`
        is a *substituent_link*

        Returns
        -------
        bool
        '''
        return (self.child.node_type is SubstituentBase.node_type) and\
            (self.parent.node_type is SaccharideBase.node_type)

    def is_bridge_link(self):
        '''
        If :attr:`parent` is a |Substituent| and :attr:`child` is a |Monosaccharide|, then `self`
        is a *bridge_link*

        Returns
        -------
        bool
        '''
        return (self.parent.node_type is SubstituentBase.node_type) and\
            (self.child.node_type is SaccharideBase.node_type)

    def is_parent(self, mol):
        '''
        Test `mol` for :func:`id` equality with :attr:`parent`

        Returns
        -------
        bool
        '''
        return (mol) is (self.parent)

    def is_child(self, mol):
        '''
        Test `mol` for :func:`id` equality with :attr:`child`

        Returns
        -------
        bool
        '''
        return (mol) is (self.child)

    def clone(self, parent, child, prop_id=True, attach=True):
        '''
        Given two new objects `parent` and `child`, duplicate all other information in `self`,
        creating a new `Link` object between `parent` and `child` with the same properties as
        `self`.

        Returns
        -------
        Link
        '''
        return Link(parent, child,
                    parent_position=self.parent_position,
                    child_position=self.child_position,
                    parent_loss=self.parent_loss,
                    child_loss=self.child_loss,
                    id=self.id if prop_id else uuid4().int,
                    attach=attach,
                    parent_linkage_type=self.parent_linkage_type,
                    child_linkage_type=self.child_linkage_type)

    def _get_parent_configuration(self):
        # Possible problem since Monosaccharide objects do not define __hash__
        return {self.parent}, {self.parent_position}

    def _get_child_configuration(self):
        # Possible problem since Monosaccharide objects do not define __hash__
        return {self.child}, {self.child_position}

    def __eq__(self, other):
        '''
        Performs deep equality testing between `self` and `other`
        '''
        if (other is None or not isinstance(other, Link)):
            return False
        if self.is_ambiguous() or other.is_ambiguous():
            res = self._get_parent_configuration() == other._get_parent_configuration()
            if res:
                res = self._get_child_configuration() == other._get_child_configuration()
            if res:
                res = self._flat_equality(other)
            if res:
                res = (self.parent.id == other.parent.id) and\
                      (self.parent == other.parent)
            return res
        else:
            res = self._flat_equality(other)
            if res:
                res = (self.parent.id == other.parent.id) and\
                      (self.parent == other.parent)
            return res

    def trait_equality(self, other):
        """Test whether the intrinsic attributes of another :class:`Link` object
        match this link without examining the actual termini.

        Parameters
        ----------
        other : :class:`Link`
            The object to compare against

        Returns
        -------
        bool
        """
        res = True
        if res:
            res = (self.parent_position == other.parent_position)
        if res:
            res = (self.child_position == other.child_position)
        if res:
            res = (self.parent_loss == other.parent_loss)
        if res:
            res = (self.child_loss == other.child_loss)
        return res

    def _flat_equality(self, other):
        res = self.trait_equality(other)
        if res:
            res = (self.child.id == other.child.id) and\
                  (self.child == other.child)
        return res

    def __ne__(self, other):
        return not (self == other)

    def break_link(self, refund=False):
        '''
        This function reverses :meth:`Link.apply`, removing the reference to :obj:`self` from
        both :attr:`parent` and :attr:`child`'s store of links. If ``refund`` is :const:`True`,
        :meth:`Link.refund` will be called, restoring the lost elemental composition from this
        bond.

        Sets :attr:`_attached` to |False|

        Returns
        -------
        :class:`~.monosaccharide.Monosaccharide` or :class:`~.substituent.Substituent` parent
        :class:`~.monosaccharide.Monosaccharide` or :class:`~.substituent.Substituent` child
        '''
        if self.is_substituent_link():
            self.parent.substituent_links.pop(self.parent_position, self)
        else:
            self.parent.links.pop(self.parent_position, self)
        self.child.links.pop(self.child_position, self)
        if refund:
            self.refund()
        self.parent._degree -= 1
        self.child._degree -= 1
        self._attached = False
        return (self.parent, self.child)

    def _reconnect(self, refund=False):
        '''
        The opposite of :meth:`Link.break_link`, add `self` to the appropriate places on
        :attr:`parent` and :attr:`child`

        Parameters
        ----------
        refund: bool
            Should :meth:`Link.refund` be called? Defaults to |False|

        '''
        if self.is_substituent_link():
            self.parent.substituent_links[self.parent_position] = self
        else:
            self.parent.links[self.parent_position] = self

        self.child.links[self.child_position] = self

        if refund:
            self.refund()
        self.parent._degree += 1
        self.child._degree += 1
        self._attached = True

    def refund(self):
        '''
        Returns the lost elemental composition caused by :meth:`apply`. Adds back :attr:`parent_loss`
        and :attr:`child_loss` to the :attr:`composition` of :attr:`parent` and :attr:`child` respectively
        '''
        self.parent.composition += (self.parent_loss or default_parent_loss)
        self.child.composition += (self.child_loss or default_child_loss)

    def is_attached(self, deep=False):
        '''
        Test to see if `self` is present in :attr:`parent` and :attr:`child` link structures.
        Presences indicates the link is attached.

        Returns
        -------
        |bool|
        '''
        if not deep:
            return self._attached
        parent = False
        child = False
        if self.is_substituent_link():
            parent_list = self.parent.substituent_links[self.parent_position]
        else:
            parent_list = self.parent.links[self.parent_position]
        for lin in parent_list:
            if lin.id == self.id:
                if lin == self:
                    parent = True
                    break
        for lin in self.child.links[self.child_position]:
            if lin.id == self.id:
                if lin == self:
                    child = True
                    break
        assert self._attached == (parent and child)
        res = self._attached = (parent and child)
        return res

    def try_break_link(self, refund=False):  # pragma: no cover
        '''
        Try to break the link if it is attached, otherwise return |False|
        '''
        if self.is_attached():
            return self.break_link(refund=refund)
        return False

    def try_apply(self):  # pragma: no cover
        '''
        Try to apply the link if it is not attached, otherwise return |False|
        '''
        if not self.is_attached():
            return self.apply()
        return False

    def _glycoct_sigils(self):
        '''
        Helper method for determining which GlycoCT symbols and losses to present
        '''
        parent_loss_str = 'x'
        child_loss_str = 'x'
        water = Composition({"O": 1, "H": 1})

        if self.child_loss == water:
            child_loss_str = "d"
            parent_loss_str = "o"
        elif self.parent_loss == water:
            child_loss_str = 'o'
            parent_loss_str = 'd'

        if self.child_loss == Composition(
                H=1) and (self.child.node_type is SubstituentBase.node_type):
            child_loss_str = "n"
            if self.parent_loss == water:
                parent_loss_str = "d"
            else:
                parent_loss_str = "o"

        if self.child_loss is None:
            child_loss_str = 'x'
        if self.parent_loss is None:
            parent_loss_str = 'x'

        return parent_loss_str, child_loss_str

    def __repr__(self):  # pragma: no cover
        parent_loss_str, child_loss_str = self._glycoct_sigils()
        rep = "({parent.id}){parent_loss}({parent_position}+{child_position}){child_loss}({child.id})"
        rep = rep.format(
            parent=self.parent,
            parent_loss=parent_loss_str,
            parent_position=self.parent_position,
            child=self.child,
            child_loss=child_loss_str,
            child_position=self.child_position)
        if self.is_attached():
            rep += "[x]"
        else:
            rep += "[ ]"
        return rep

    def __getstate__(self):
        state = {}
        state["parent"] = self.parent
        state["child"] = self.child
        state["parent_position"] = self.parent_position
        state["child_position"] = self.child_position
        state["parent_loss"] = self.parent_loss
        state["child_loss"] = self.child_loss
        state["id"] = self.id
        state["label"] = self.label
        state["_attached"] = self._attached
        state['parent_linkage_type'] = self.parent_linkage_type
        state['child_linkage_type'] = self.child_linkage_type
        return state

    def __setstate__(self, state):
        self.parent = state["parent"]
        self.child = state["child"]
        self.parent_position = state["parent_position"]
        self.child_position = state["child_position"]
        self.parent_loss = state["parent_loss"]
        self.child_loss = state["child_loss"]
        self.id = state["id"]
        self.label = state["label"]
        self._attached = state["_attached"]
        try:
            self.parent_linkage_type = state['parent_linkage_type']
        except KeyError:
            self.parent_linkage_type = LinkageType.unknown
        try:
            self.child_linkage_type = state['child_linkage_type']
        except KeyError:
            self.child_linkage_type = LinkageType.unknown


class LinkMaskContext(object):
    '''
    A context manager for masking and unmasking |Link| objects on a residue
    '''
    def __init__(self, residue, attach=False):
        self.residue = residue
        self.links = [link for link in residue.links.values()]
        self.attach = attach

    def __enter__(self):
        self.mask() if not self.attach else self.unmask()

    def __exit__(self, exc_type, exc_value, traceback):
        self.mask() if self.attach else self.unmask()

    def mask(self):
        """
        For each |Link| captured, if that link is attched, call its
        :meth:`break_link` method with `refund=True`
        """
        for link in self.links:
            if link.is_attached():
                link.break_link(refund=True)

    def unmask(self):
        """
        For each |Link| captured, if that link is not attched, call its
        :meth:`apply` method
        """
        for link in self.links:
            if not link.is_attached():
                link.apply()


class AmbiguousLink(Link):
    '''A :class:`Link` that may have more than one option for connection position at either
    terminal, or the terminals themselves may be chosen from a set of options.

    Attributes
    ----------
    parent_choices: list
        The set of possible choices for :attr:`parent`
    parent_position_choices: list
        The set of possible choices for :attr:`parent_position`
    child_choices: list
        The set of possible choices for :attr:`child`
    child_position_choices: list
        The set of possible choices for :attr:`child_position`
    '''

    __slots__ = (
        "parent_choices", "parent_position_choices",
        "child_choices", "child_position_choices"
    )

    def __init__(self, parent, child, parent_position=(UnknownPosition,), child_position=(UnknownPosition,),
                 parent_loss=None, child_loss=None, id=None, attach=True, parent_linkage_type=None,
                 child_linkage_type=None):
        if not isinstance(parent, (list, tuple)):
            parent = [parent]
        if not isinstance(child, (list, tuple)):
            child = [child]
        if not isinstance(parent_position, Iterable):
            parent_position = [parent_position]
        if not isinstance(child_position, Iterable):
            child_position = [child_position]

        self.parent_choices = list(parent)
        self.parent_position_choices = list(parent_position)

        self.child_choices = list(child)
        self.child_position_choices = list(child_position)

        super(AmbiguousLink, self).__init__(
            self.parent_choices[0],
            self.child_choices[0],
            self.parent_position_choices[0],
            self.child_position_choices[0],
            parent_loss, child_loss, id, attach,
            parent_linkage_type, child_linkage_type)

    def __getstate__(self):
        state = super(AmbiguousLink, self).__getstate__()
        state['parent_choices'] = self.parent_choices
        state['parent_position_choices'] = self.parent_position_choices
        state['child_choices'] = self.child_choices
        state['child_position_choices'] = self.child_position_choices
        return state

    def __setstate__(self, state):
        super(AmbiguousLink, self).__setstate__(state)
        self.parent_choices = state['parent_choices']
        self.parent_position_choices = state['parent_position_choices']
        self.child_choices = state['child_choices']
        self.child_position_choices = state['child_position_choices']

    def has_ambiguous_linkage(self):
        return len(self.parent_position_choices) > 1 or len(self.child_position_choices) > 1

    def has_ambiguous_termini(self):
        return len(self.parent_choices) > 1 or len(self.child_choices) > 1

    def iterconfiguration(self, attach=False):
        """Iterate over possible configurations for this link, yielding them.

        If ``attach`` is :const:`True`, call :meth:`reconfigure` with each
        specification.

        Parameters
        ----------
        attach: :class:`bool`, optional
            Whether or not to apply each configuration. The default is :const:`False`

        Yields
        ------
        :class:`tuple` of :attr:`parent`, :attr:`child`, :attr:`parent_position`, :attr:`child_position`
        """
        configurations = self._configuration_crossproduct()
        for parent_o, child_o, parent_position_o, child_position_o in configurations:
            if attach:
                self.reconfigure(parent_o, child_o, parent_position_o, child_position_o)
            yield linkage_configuration(parent_o, child_o, parent_position_o, child_position_o)

    def _configuration_crossproduct(self):
        configurations = itertools.product(self.parent_choices, self.child_choices,
                                           self.parent_position_choices,
                                           self.child_position_choices)
        return configurations

    def reconfigure(self, parent, child, parent_position, child_position):
        """Apply the specified configuration, replacing the current configuration.

        This method calls :meth:`break_link` to refund the current bonds and
        then calls :meth:`apply`.

        Parameters
        ----------
        parent : :class:`~.Monosaccharide` or :class:`~.Substituent`
            The parent of the link to form
        child : :class:`~.Monosaccharide` or :class:`~.Substituent
            The child of the link to form
        parent_position : int
            The position on the parent to connect to
        child_position : int
            The position on the child to connect to
        """
        self.break_link(refund=True)
        self.parent = parent
        self.child = child
        self.parent_position = parent_position
        self.child_position = child_position
        self.apply()

    def clone(self, parent, child, prop_id=True, attach=True):
        if not isinstance(parent, (list, tuple)):
            parent = [parent]
        if not isinstance(child, (list, tuple)):
            child = [child]
        link = AmbiguousLink(
            parent, child,
            self.parent_position_choices,
            self.child_position_choices,
            self.parent_loss, self.child_loss, self.id if prop_id else None,
            attach=False)
        link.parent_position = self.parent_position
        link.child_position = self.child_position
        if attach:
            link.apply()

        return link

    def break_link(self, refund=True):
        res = super(AmbiguousLink, self).break_link(refund=refund)
        assert not self.is_attached(deep=True)
        return res

    def __repr__(self):
        rep = super(AmbiguousLink, self).__repr__()
        return "(A)" + rep

    def _find_open_position_single(self, parent, child):
        open_parent_sites, _ = parent.open_attachment_sites()
        parent_has_unknown_position = False
        if UnknownPosition in open_parent_sites:
            open_parent_sites += list(set(self.parent_position_choices) - set(parent.occupied_attachment_sites()))
            parent_has_unknown_position = True
        parent_site_options = list((set(open_parent_sites) | {UnknownPosition}) & set(self.parent_position_choices))
        if parent_site_options:
            parent_site = parent_site_options[0]
        else:
            if parent_has_unknown_position and open_parent_sites:
                parent_site = UnknownPosition
            else:
                return None
        open_child_sites, _ = child.open_attachment_sites()
        child_has_unknown_position = False
        if UnknownPosition in open_child_sites:
            open_child_sites += list(set(self.child_position_choices) - set(child.occupied_attachment_sites()))
            child_has_unknown_position = True
        child_site_options = list((set(open_child_sites) | {UnknownPosition}) & set(self.child_position_choices))
        if child_site_options:
            child_site = child_site_options[0]
        else:
            if child_has_unknown_position and open_child_sites:
                child_site = UnknownPosition
            else:
                return None
        return parent_site, child_site

    def _find_open_position_multiple(self, parent, child):  # pragma: no cover
        '''Debugging purposes only
        '''
        open_parent_sites, _ = parent.open_attachment_sites()
        if UnknownPosition in open_parent_sites:
            open_parent_sites += self.parent_position_choices
        parent_site_options = list((set(open_parent_sites) | {UnknownPosition}) & set(self.parent_position_choices))
        open_child_sites, _ = child.open_attachment_sites()
        if UnknownPosition in open_child_sites:
            open_child_sites += self.child_position_choices
        child_site_options = list((set(open_child_sites) | {UnknownPosition}) & set(self.child_position_choices))
        return parent_site_options, child_site_options

    def _find_open_position_combinations(self):
        for parent, child in itertools.product(self.parent_choices, self.child_choices):
            positions = self._find_open_position_single(parent, child)
            if positions is not None:
                return (parent, positions[0], child, positions[1])

    def find_open_position(self, attach=True):
        """Identify a valid configuration of parent, child and respective positions
        which are all open.

        Parameters
        ----------
        attach : bool, optional
            Whether or not to actually connect to the selected configuration, which will
            break the current configuration (the default is :const:`True`).

        Raises
        ------
        ValueError:
            When no valid configuration can be found.

        """
        if attach:
            self.break_link(refund=True)
        configuration = self._find_open_position_combinations()
        if configuration is None:
            raise ValueError("Could not find a valid configurations on current parent/child pair")
        parent, parent_position, child, child_position = configuration
        self.parent = parent
        self.child = child
        self.parent_position = parent_position
        self.child_position = child_position
        if attach:
            self.apply()



import warnings

from collections import defaultdict

from glypy.composition import Composition

do_warn = True
do_error = False


class UnknownCompositionWarning(Warning):
    pass


class CompositionRule(object):
    def __init__(self, base_composition, position_shifts=None):
        if position_shifts is None:
            position_shifts = {}
        self.composition = Composition(base_composition)
        self.position_shifts = defaultdict(Composition)
        self.position_shifts.update(position_shifts)

    def __call__(self, position=-1):
        # Implicitly copies
        return self.composition + Composition(self.position_shifts[position])


class CompositionIndex(dict):

    def __getitem__(self, key):
        try:
            composition_dict = super(CompositionIndex, self).__getitem__(key)
        except KeyError:
            composition_dict = Composition()
            if do_warn:
                warnings.warn("{key} could not be found. It may not have an explicit composition".format(key=key),
                              UnknownCompositionWarning,
                              stacklevel=3)
            if do_error:
                raise
        # Explicitly copies
        return (composition_dict).clone()

    def __setitem__(self, key, value):
        value = Composition(value)
        super(CompositionIndex, self).__setitem__(key, value)

    def register(self, key, value):
        self[key] = value


class CompositionRuleIndex(dict):

    def __getitem__(self, key):
        try:
            rule = super(CompositionRuleIndex, self).__getitem__(key)
        except KeyError:  # pragma: no cover
            rule = CompositionRule(Composition())
            if do_warn:
                warnings.warn("{key} could not be found. It may not have an explicit composition".format(key=key),
                              stacklevel=3)
        return (rule)


# Monosaccharide Residue Compositions
_monosaccharide_compositions = {
    "sug": Composition({
        "C": 2,
        "H": 4,
        "O": 2
    }),
    "tri": Composition({
        "C": 3,
        "H": 6,
        "O": 3
    }),
    "tet": Composition({
        "C": 4,
        "H": 8,
        "O": 4
    }),
    "pen": Composition({
        "C": 5,
        "H": 10,
        "O": 5
    }),
    "hex": Composition({
        "C": 6,
        "H": 12,
        "O": 6
    }),
    "hep": Composition({
        "C": 7,
        "H": 14,
        "O": 7
    }),
    "oct": Composition({
        "C": 8,
        "H": 16,
        "O": 8
    }),
    "non": Composition({
        "C": 9,
        "H": 18,
        "O": 9
    }),
    "dec": Composition({
        "C": 10,
        "H": 20,
        "O": 10
    }),
    "s11": Composition("H2CO") * 11,
    "s12": Composition("H2CO") * 12,
    "s13": Composition("H2CO") * 13,
    "s14": Composition("H2CO") * 14,
    "s15": Composition("H2CO") * 15,
    "s16": Composition("H2CO") * 16,
    "s17": Composition("H2CO") * 17,
    "s18": Composition("H2CO") * 18,
    "s19": Composition("H2CO") * 19,
    "s20": Composition("H2CO") * 20
}

monosaccharide_composition = CompositionIndex(_monosaccharide_compositions)

#: Modification Compositions defined as dictionaries, specified as composition
#: deltas relative to the glycan structure
_modification_compositions = {
    "d": CompositionRule({
        "O": -1,
    }),
    "a": CompositionRule({
        "H": -2,
        "O": 1
    }),
    "keto": CompositionRule({"H": -2}),
    "en": CompositionRule({
        "H": -2,
        "O": -1
    })
}

modification_compositions = CompositionRuleIndex(_modification_compositions)

# Specified as composition deltas relative to the glycan structure
_substituent_compositions = {
    "n_acetyl": Composition({
        "C": 2,
        "H": 5,
        "N": 1,
        "O": 1,
    }),
    "sulfate": Composition({
        "S": 1,
        "O": 3,
        "H": 2
    }),
    "amino": Composition({
        "N": 1,
        "H": 3,
    }),
    "n_glycolyl": Composition({
        "C": 2,
        "H": 5,
        "O": 2,
        "N": 1
    }),
    "phosphate": Composition({
        'P': 1,
        'O': 3,
        'H': 3
    }),
    "acetyl": Composition({
        'H': 4,
        'C': 2,
        'O': 1
    }),
    "methyl": Composition({
        'H': 4,
        'C': 1
    }),
    "n_sulfate": Composition({
        "H": 3,
        'S': 1,
        'O': 3,
        'N': 1
    }),
    "nitrate": Composition({
        "N": 1,
        "O": 3
    }),
    "iodo": Composition({
        "I": 1,
        "H": 1,
    }),
    "fluoro": Composition("FH"),
    "imino": Composition("NH2"),
    "chloro": Composition("ClH"),
    "formyl": Composition("CHOH"),
    "bromo": Composition("BrH"),
    "n_alanine": Composition({
        "C": 3,
        "H": 8,
        "N": 2,
        "O": 1
    }),
    "ethyl": Composition("CH2CH3H"),
    'n_ethyl': Composition("NHCH2CH3H"),
    "n_dimethyl": Composition("N(CH3)2H"),
    "n_formyl": Composition("NHCHOH"),
    "n_methyl": Composition("NHCH3H"),
    "dimethylamine": Composition("NHC2H6"),
    "n_succinate": Composition("NCOCH2CH2COOHH"),
    "n_trifluoroacetyl": Composition("NHCOCF3H"),
    "thio": Composition("SHH"),
    "(r)_pyruvate": Composition("CH2CCOOH"),
    "(s)_pyruvate": Composition("CH2CCOOH"),
    "pyruvate": Composition("COCOCH3H"),
    "(r)_lactate": Composition("CH3CHCOOH2"),
    "(s)_lactate": Composition("CH3CHCOOH2"),
    "ethanolamine": Composition("NHCH2CH2OHH"),
    "lactonization": -Composition("O"),
    "anhydro": -Composition("O"),  # Semantically, anhydro is more like a Modification
    "phospho_ethanolamine": Composition("C2H8NO3P"),
    "glycolyl": Composition("COCH2OH"),
    "esterification": Composition("CH3CH3")
}

substituent_compositions = CompositionIndex(_substituent_compositions)



'''
A parser for :title-reference:`GlycoCT{XML}` format.

This module provides a simple parser for GlycoCT XML. It has fewer
features than :mod:`glypy.io.glycoct`, and does not support writing.

It is included for historical purposes.
'''
# pragma: no cover

from operator import itemgetter
from collections import defaultdict
from glypy.utils import opener, StringIO, ET
from glypy.utils.multimap import OrderedMultiMap
from glypy.structure import monosaccharide, substituent, link
from glypy.structure.glycan import Glycan
from .format_constants_map import (anomer_map, superclass_map,
                                   link_replacement_composition_map, modification_map)
from .tree_builder_utils import try_int
from .file_utils import ParserError

basetype_unpacker = itemgetter("id", "anomer", "superclass", "ringStart", "ringEnd")

START = "!START"
RES = "RES"
LIN = "LIN"
REP = "REP"
ALT = "ALT"
UND = "UND"


class GlycoCTXMLError(ParserError):
    pass


class GlycoCTXMLSectionUnsupported(GlycoCTXMLError):
    pass


class GlycoCTXML(object):
    '''Parse :title-reference:`GlycoCT{XML}` text data into |Glycan| objects.

    The parser implements the :class:`Iterator` interface, yielding successive glycans
    from a text stream separated by empty lines.
    '''
    @classmethod
    def loads(cls, glycoct_str):
        '''Parse results from |str|'''
        rep = StringIO(glycoct_str)
        return cls(rep)

    def __init__(self, stream, structure_class=Glycan):
        self.graph = {}
        self.state = START
        self.handle = opener(stream, "r")
        self.counter = 0
        self.repeats = {}
        self.buffer = defaultdict(list)
        self.root = None
        self._iter = None
        self.structure_class = structure_class

    def _reset(self):
        self.graph = {}
        self.buffer = defaultdict(list)
        self.root = None
        self.state = START

    def __iter__(self):
        '''
        Calls :meth:`parse` and stores it for reuse with :meth:`__next__`
        '''
        self._iter = self.parse()
        return self._iter

    def next(self):
        '''
        Calls :meth:`parse` if the internal iterator has not been instantiated

        '''
        if self._iter is None:
            iter(self)
        return next(self._iter)

    __next__ = next

    def _make_iterator(self):
        for evt, entity in ET.iterparse(self.handle, ("start", "end")):
            entity.tag = entity.tag.split("}")[-1]
            yield evt, entity

    def parse(self):
        for evt, entity in self._make_iterator():
            if evt != "end":
                if entity.tag == "sugar":
                    self.state = START
                elif entity.tag == 'residues':
                    if self.state == START:
                        self.state = RES
                    else:
                        raise GlycoCTXMLError("<residues> not the first section encountered")
                elif entity.tag == 'linkages':
                    if self.state != RES:
                        raise GlycoCTXMLError("<linkages> not following <residues>")
                elif entity.tag in {'repeat'}:
                    raise GlycoCTXMLSectionUnsupported("<{}> section is not supported".format(entity.tag))
                continue
            else:
                if entity.tag == "stemtype":
                    self.buffer['stem'].append(entity.attrib['type'][1:])
                    self.buffer['configuration'].append(entity.attrib['type'][0])
                elif entity.tag == "modification":
                    self.buffer['modification'].append(
                        (entity.attrib['type'], try_int(entity.attrib['pos_one'])))
                elif entity.tag == "basetype":
                    id, anomer, superclass, ring_start, ring_end = basetype_unpacker(entity.attrib)
                    superclass = superclass_map[superclass.upper()]
                    anomer = anomer_map[anomer]
                    id = int(id)
                    modifications = OrderedMultiMap()
                    mods = self.buffer.pop("modification", None)
                    if mods is not None:
                        for mod, pos in mods:
                            modifications[pos] = modification_map[mod]
                    is_reduced = "aldi" in modifications[1]
                    if is_reduced:
                        modifications.pop(1, "aldi")

                    residue = monosaccharide.Monosaccharide(
                        anomer=anomer, superclass=superclass, stem=self.buffer.pop('stem'),
                        configuration=self.buffer.pop("configuration"), ring_start=ring_start,
                        ring_end=ring_end, modifications=modifications, reduced=is_reduced, id=id)
                    self.graph[id] = residue
                    if self.root is None:
                        self.root = residue
                elif entity.tag == "substituent":
                    substituent_obj = substituent.Substituent(entity.attrib['name'])
                    self.graph[int(entity.attrib['id'])] = substituent_obj
                elif entity.tag == "connection":
                    parent_id = try_int(entity.attrib['parent'])
                    child_id = try_int(entity.attrib['child'])
                    parent_node = self.graph[parent_id]
                    child_node = self.graph[child_id]
                    link.Link(
                        parent_node, child_node, parent_position=self.buffer.pop("parent_position"),
                        child_position=self.buffer.pop("child_position"),
                        parent_loss=self.buffer.pop('parent_loss'), child_loss=self.buffer.pop('child_loss'),
                        id=self.buffer.pop('id'))
                elif entity.tag == "parent":
                    self.buffer['parent_position'] = int(entity.attrib['pos'])
                elif entity.tag == "child":
                    self.buffer['child_position'] = int(entity.attrib['pos'])
                elif entity.tag == "linkage":
                    self.buffer["id"] = int(entity.attrib['id'])
                    self.buffer["parent_loss"] = link_replacement_composition_map[entity.attrib['parentType']]
                    self.buffer["child_loss"] = link_replacement_composition_map[entity.attrib['childType']]
                elif entity.tag == 'sugar':
                    if self.root is not None:
                        yield self.structure_class(self.root)
                    self._reset()
                entity.clear()


def read(stream):
    '''
    A convenience wrapper for :class:`GlycoCTXML`

    Parameters
    ----------
    stream : file-like
        The text stream to parse structures from

    Returns
    -------
    :class:`~.GlycoCTXML`
    '''
    return GlycoCTXML(stream)


def load(stream, structure_class=Glycan, allow_multiple=True):
    """Read all structures from the provided text stream.

    Parameters
    ----------
    stream : file-like
        The text stream to parse structures from
    structure_class : type, optional
        :class:`~.Glycan` subclass to use
    allow_multiple: bool, optional
        If :const:`False`, makes loading multiple structures an error.

    Returns
    -------
    :class:`~.Glycan` or :class:`list` of :class:`~.Glycan`
    """

    g = GlycoCTXML(stream, structure_class=structure_class)
    first = next(g)
    if not allow_multiple:
        return first
    second = None
    try:
        second = next(g)
        collection = [first, second]
        collection.extend(g)
        return collection
    except StopIteration:
        return first


def loads(text, allow_multiple=True):
    """Read all structures from the provided text string.

    Parameters
    ----------
    text : str
        The text to parse structures from

    Returns
    -------
    :class:`~.Glycan` or :class:`list` of :class:`~.Glycan`
    """

    g = GlycoCTXML.loads(text)
    first = next(g)
    if not allow_multiple:
        return first
    second = None
    try:
        second = next(g)
        collection = [first, second]
        collection.extend(g)
        return collection
    except StopIteration:
        return first



from glypy.composition.structure_composition import substituent_compositions

from .base import SubstituentBase
from .link import Link
from .constants import UnknownPosition

from glypy.utils import uid
from glypy.composition import Composition, calculate_mass
from glypy.utils.multimap import OrderedMultiMap


class DerivatizePathway(object):

    def __init__(self, can_nh_derivatize=False, is_nh_derivatizable=False, name=None):
        self.can_nh_derivatize = can_nh_derivatize
        self.is_nh_derivatizable = is_nh_derivatizable
        if name is not None:
            derivatize_info[name] = self

    def __repr__(self):  # pragma: no cover
        return "<DerivatizePathway {}>".format(self.__dict__)

    @classmethod
    def register(cls, name, can_nh_derivatize, is_nh_derivatizable):
        derivatize_info[Substituent.internalize_name(name)] = DerivatizePathway(can_nh_derivatize, is_nh_derivatizable)


attachment_composition_info = {
    "sulfate": Composition("H"),
    "methyl": Composition("H"),
    "n_acetyl": Composition("OH"),
    "n_glycolyl": Composition("OH"),
    "n_sulfate": Composition("OH"),
    "amino": Composition("OH"),
    "imino": Composition("OH"),
    "anhydro": Composition("H"),
    "dimethylamine": Composition("OH"),
    "phosphate": Composition("H")
}
default_attachment_composition = Composition("H")


derivatize_info = {
    "acetyl": DerivatizePathway(False, False),
    "amino": DerivatizePathway(True, False),
    "anhydro": DerivatizePathway(True, False),
    "bromo": DerivatizePathway(True, False),
    "chloro": DerivatizePathway(True, False),
    "diphospho_ethanolamine": DerivatizePathway(True, False),
    "ethanolamine": DerivatizePathway(True, False),
    "ethyl": DerivatizePathway(True, False),
    "fluoro": DerivatizePathway(True, False),
    "formyl": DerivatizePathway(True, False),
    "glycolyl": DerivatizePathway(True, False),
    "hydroxymethyl": DerivatizePathway(True, False),
    "iodo": DerivatizePathway(True, False),
    "lactone": DerivatizePathway(True, False),
    "methyl": DerivatizePathway(True, False),
    "phosphate": DerivatizePathway(True, False),
    "phosphocholine": DerivatizePathway(True, False),
    "phospho_ethanolamine": DerivatizePathway(True, False),
    "pyrophosphate": DerivatizePathway(True, False),
    "pyruvate": DerivatizePathway(True, False),
    "succinate": DerivatizePathway(True, False),
    "sulfate": DerivatizePathway(True, False),
    "thio": DerivatizePathway(True, False),
    "triphosphate": DerivatizePathway(True, False),

    "(x)_lactate": DerivatizePathway(True, False),
    "(r)_carboxyethyl": DerivatizePathway(True, False),
    "(r)_lactate": DerivatizePathway(True, False),
    "(r)_pyruvate": DerivatizePathway(True, False),
    "(s)_carboxyethyl": DerivatizePathway(True, False),
    "(s)_lactate": DerivatizePathway(True, False),
    "(s)_pyruvate": DerivatizePathway(True, False),

    "n_acetyl": DerivatizePathway(True, True),
    "n_amidino": DerivatizePathway(True, True),
    "n_formyl": DerivatizePathway(True, True),
    "n_glycolyl": DerivatizePathway(True, True),
    "n_methyl": DerivatizePathway(True, True),
    "n_succinate": DerivatizePathway(True, True),
    "n_sulfate": DerivatizePathway(True, True),
    "n_dimethyl": DerivatizePathway(True, True),

    "phospho_choline": DerivatizePathway(True, False),
}


def register(name, composition, can_nh_derivatize=None, is_nh_derivatizable=None,
             attachment_composition=None):
    """Register common information about a |Substituent| group to be used during initialization
    of instances of |Substituent| which share that name.

    Parameters
    ----------
    name : str
        The name to be registered
    composition : |Composition|
        The shared base composition that will be initialized for each instance
    can_nh_derivatize : None, optional
        Passed to `DerivatizePathway.register`
    is_nh_derivatizable : None, optional
        Passed to `DerivatizePathway.register`
    attachment_composition : None, optional
        The shared composition that will be lost from the parent molecule when forming
        a bond with substituents of this type.
    """
    name = Substituent.internalize_name(name)
    substituent_compositions[name] = composition.clone()
    attachment_composition_info[name] = attachment_composition if attachment_composition is not None\
        else default_attachment_composition
    DerivatizePathway.register(name, can_nh_derivatize or False, is_nh_derivatizable or False)
    return Substituent(name)


def unregister(name):
    """Removes all information about the |Substituent| group denoted by `name` from
    the shared indices.

    Parameters
    ----------
    name : str
        The name to un-register
    """
    name = Substituent.internalize_name(name)
    substituent_compositions.pop(name)
    attachment_composition_info.pop(name)
    derivatize_info.pop(name)


def is_registered(name):
    name = Substituent.internalize_name(name)
    return name in substituent_compositions


class Substituent(SubstituentBase):
    '''
    Represents a non-saccharide molecule commonly found bound to saccharide units.

    Attributes
    ----------
    name: |str|
        The name of the substituent, used to uniquely identify it.
    links: |OrderedMultiMap|
        All links to all molecules connected to this one.
    composition: |Composition|
        The chemical makeup of this molecule.
    attachment_composition: |Composition|
        The default cost of attaching this substituent to a |Monosaccharide|
    id: |int|
        A unique identifier number for this molecule.

    can_nh_derivatize: |bool|
        Whether this substituent will derivatize at an amine group.
    is_nh_derivatizable: |bool|
        Whether this substituent contains a derivatizable amine group.
    _derivatize: |bool|
        Whether this substituent was added by a derivatization process.

    _degree: |int|
        The number of connections to this molecule. Mutated internally by |Link| objects,
        not for external use. See :meth:`order`.

    '''
    register = staticmethod(register)
    unregister = staticmethod(unregister)

    __slots__ = (
        "_name", "links", "composition", "id",
        "can_nh_derivatize", "is_nh_derivatizable",
        "_derivatize", "attachment_composition",
        "_degree"
    )

    def __init__(self, name, links=None, composition=None, id=None,
                 can_nh_derivatize=None, is_nh_derivatizable=None, derivatize=False,
                 attachment_composition=None):
        if links is None:
            links = OrderedMultiMap()
        self.name = name
        self.links = links
        if composition is None:
            composition = substituent_compositions[self._name]
        elif composition is not None and not is_registered(self._name):
            self.register(
                name, composition, can_nh_derivatize=can_nh_derivatize,
                is_nh_derivatizable=is_nh_derivatizable,
                attachment_composition=attachment_composition)

        self.composition = composition
        self.id = id or uid()
        self._degree = self.order()
        try:
            if can_nh_derivatize is is_nh_derivatizable is None:
                derivatize_pathway = derivatize_info[self.name]
                self.can_nh_derivatize = derivatize_pathway.can_nh_derivatize
                self.is_nh_derivatizable = derivatize_pathway.is_nh_derivatizable
            else:
                self.can_nh_derivatize = can_nh_derivatize or False
                self.is_nh_derivatizable = is_nh_derivatizable or False
        except KeyError:
            self.can_nh_derivatize = can_nh_derivatize or False
            self.is_nh_derivatizable = is_nh_derivatizable or False
        self._derivatize = derivatize
        self.attachment_composition = attachment_composition if attachment_composition is not None\
            else attachment_composition_info.get(self.name, default_attachment_composition)

    @staticmethod
    def internalize_name(name):
        return name.replace('-', '_')

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        '''
        Translate the name of the substituent from the common dash-separated notation
        to a valid identifier, replacing - with _.

        Parameters
        ----------
        value: str
            The name being set

        '''
        self._name = self.internalize_name(value)

    def __repr__(self):  # pragma: no cover
        return "<Substituent {name}>".format(name=self._name)

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return (other is not None) and (self.name == other.name) and (self.composition == other.composition)

    def __ne__(self, other):
        return not self == other

    def open_attachment_sites(self):
        return [1, 2], 0

    def is_occupied(self, position):
        '''
        Check to see if `position` is occupied. Unlike |Monosaccharide|, |Substituent| objects
        can only have two attachment sites at this time.

        Parameters
        ----------
        position: int

        Returns
        -------
        int:
            Number of links at `position`

        Raises
        ------
        IndexError:
            If `position` > 2 or < 1
        '''
        if position > 2 or position < 1:
            raise IndexError("Position out of range")
        return len(self.links[position])

    def add_substituent(self, substitent, position=2, max_occupancy=1,
                        child_position=1, parent_loss=None, child_loss=None):
        '''
        Adds a :class:`~glypy.structure.substituent.Substituent` and associated :class:`~glypy.structure.link.Link`
        to :attr:`links` at the site given by ``position``. This new substituent is included when calculating mass
        with substituents included

        Parameters
        ----------
        substituent: str or Substituent
            The substituent to add. If passed a |str|, it will be
            translated into an instance of |Substituent|
        position: int or 'x'
            The location to add the |Substituent| link to :attr:`links`. Defaults to 2
        child_position: int
            The location to add the link to in `substituent`'s :attr:`links`. Defaults to 1. Substituent
            indices are currently not checked.
        max_occupancy: int, optional
            The maximum number of items acceptable at ``position``. Defaults to :const:`1`
        parent_loss: Composition or str
            The elemental composition removed from ``self``. Defaults to ``H1``.
        child_loss: Composition or str
            The elemental composition removed from ``substituent``. Defaults to ``H1``.

        Raises
        ------
        ValueError:
            ``position`` is occupied by more than ``max_occupancy`` elements
        '''

        if self.is_occupied(position) > max_occupancy:
            raise ValueError("Site is already occupied")
        if parent_loss is None:
            parent_loss = Composition(H=1)
        if child_loss is None:
            child_loss = Composition(H=1)
        Link(parent=self, child=substitent,
             parent_position=position, child_position=child_position,
             parent_loss=parent_loss, child_loss=child_loss)
        return self

    def drop_substituent(self, position, substituent=None, refund=True):
        link_obj = None
        for substituent_link in self.links[position]:
            if substituent_link.child == substituent or substituent is None:
                link_obj = substituent_link
                break
        if link_obj is None:
            raise IndexError(
                "No matching substituent found at {position}".format(position=position))

        link_obj.break_link(refund=refund)
        return self

    def mass(self, average=False, charge=0, mass_data=None):
        '''
        Calculates the total mass of `self` and all nodes returned by :meth:`children`.

        Parameters
        ----------
        average: bool, optional, defaults to False
            Whether or not to use the average isotopic composition when calculating masses.
            When ``average == False``, masses are calculated using monoisotopic mass.
        charge: int, optional, defaults to 0
            If charge is non-zero, m/z is calculated, where m is the theoretical mass, and z is `charge`
        mass_data: dict, optional, defaults to `None`
            If mass_data is None, standard NIST mass and isotopic abundance data are used. Otherwise the
            contents of mass_data are assumed to contain elemental mass and isotopic abundance information.

        Returns
        -------
        :class:`float`

        See also
        --------
        :func:`glypy.composition.composition.calculate_mass`
        '''
        if charge == 0:
            mass = calculate_mass(
                self.composition, average=average, charge=0, mass_data=mass_data)
            for link_pos, child in self.children():
                mass += child.mass(average=average,
                                   charge=0, mass_data=mass_data)
        else:
            mass = self.total_composition().calc_mass(average=average, charge=charge, mass_data=mass_data)
        return mass

    def __getstate__(self):
        state = {
            "_name": self._name,
            "links": self.links,
            "composition": self.composition,
            "id": self.id,
            "can_nh_derivatize": self.can_nh_derivatize,
            "is_nh_derivatizable": self.is_nh_derivatizable,
            "_derivatize": self._derivatize,
            "attachment_composition": self.attachment_composition,
            "_degree": self._degree
        }
        return state

    def __setstate__(self, state):
        self._name = state['_name']
        self.links = state['links']
        self.composition = state['composition']
        self.id = state['id']
        self.can_nh_derivatize = state['can_nh_derivatize']
        self.is_nh_derivatizable = state['is_nh_derivatizable']
        self._derivatize = state['_derivatize']
        self.attachment_composition = state['attachment_composition']
        self._degree = state.get("_degree", len(self.links))

    def clone(self, prop_id=True):
        '''
        Duplicates this |Substituent| object, recursively copying all children
        as well.

        Parameters
        ----------
        prop_id: bool
            Whether or not to propagate :attr:`id` to the clone.


        Returns
        -------
        Substituent

        See Also
        --------
        :meth:`.structure.Monosaccharide.clone`
        '''

        substituent = self.__class__(
            self.name,
            can_nh_derivatize=self.can_nh_derivatize,
            is_nh_derivatizable=self.is_nh_derivatizable,
            id=self.id if prop_id else None,
            derivatize=self._derivatize,
            attachment_composition=self.attachment_composition)
        for pos, link in self.links.items():
            if link.is_child(self):
                continue
            sub = link.to(self)
            dup = sub.clone(prop_id=prop_id)
            link.clone(substituent, dup)
        return substituent

    def degree(self):
        return len(self.links)

    def total_composition(self):
        '''
        Computes the sum of the composition of `self` and each of its linked
        :class:`~.substituent.Substituent`s

        Returns
        -------
        :class:`~glypy.composition.Composition`
        '''
        comp = self.composition
        for pos, sub in self.children():
            comp = comp + sub.total_composition()
        return comp

    def children(self, links=False, bridging=False):
        '''
        Returns an iterator over the :class:`Monosaccharide`s which are considered
        the descendants of ``self``.
        '''
        result = []
        for pos, link in self.links.items():
            if link.is_child(self):
                continue
            if links:
                if bridging and not link.is_bridge_link():
                    continue
                result.append((pos, link))
            else:
                if bridging and not link.is_bridge_link():
                    continue
                result.append((pos, link.child))
        return result

    def parents(self, links=False):
        '''
        Returns an iterator over the objects which are considered
        the ancestors of ``self``.
        '''
        result = []
        for pos, link in self.links.items():
            if link.is_parent(self):
                continue
            if links:
                result.append((pos, link))
            else:
                result.append((pos, link.parent))
        return result

    def is_bridge(self):
        for pos, link in self.children(links=True):
            if link.is_bridge():
                return True
        return False

    def attachment_composition_loss(self):
        return self.attachment_composition.clone()

    def _backsolve_original_composition(self):
        comp = self.composition.clone()
        has_substituent_link = 0
        for pos, link in self.links.items():
            if link.is_child(self):
                if link.is_substituent_link():
                    has_substituent_link += 1
                comp += link.child_loss
            else:
                comp += link.parent_loss
        if has_substituent_link == 0:
            comp += self.attachment_composition
        return comp

    def has_undefined_linkages(self):
        for link in self.links.values():
            if link.parent_position == UnknownPosition or link.child_position == UnknownPosition:
                return True
        return False



import pickle
import unittest

from glypy.composition import composition, composition_transform
from glypy.structure import monosaccharide

from .common import load

ReducedEnd = monosaccharide.ReducedEnd


def make_composition_suite(composition_type):
    class CompositionTests(unittest.TestCase):
        def test_derivatize_bare(self):
            permethylated_reduced_mass = 1716.9033
            glycan = load("common_glycan")
            glycan.reducing_end = ReducedEnd()
            composition_transform.derivatize(glycan, 'methyl')
            self.assertAlmostEqual(glycan.mass(), permethylated_reduced_mass, 3)

        def test_strip_derivatize(self):
            glycan = load("common_glycan")
            glycan.reducing_end = ReducedEnd()
            mass = glycan.mass()
            composition_transform.derivatize(glycan, 'methyl')
            self.assertNotEqual(mass, glycan.mass())
            composition_transform.strip_derivatization(glycan)
            self.assertAlmostEqual(glycan.mass(), mass, 3)

        def test_composition_equality(self):
            self.assertEqual(composition_type("H2O"), composition_type("H2O"))
            self.assertEqual(composition_type("H2O") * 2, composition_type("(H2O)2"))

        def test_composition_substraction(self):
            self.assertEqual(
                composition_type("NH2O") - composition_type("N"), composition_type("H2O"))
            c = composition_type("NH2O")
            c -= composition_type("N")
            self.assertEqual(c, composition_type("H2O"))
            c += {"N": 1}
            self.assertEqual(c, composition_type("NH2O"))

        def test_isotope_parsing(self):
            self.assertFalse(composition_type("O[18]") == composition_type("O"))
            self.assertAlmostEqual(composition_type("O[18]").mass, 17.999, 3)
            self.assertRaises(
                composition.ChemicalCompositionError, lambda: composition_type("O[18.5]"))

        def test_inits(self):
            self.assertEqual(composition_type(O=1, H=2), composition_type(formula='H2O'))
            self.assertEqual(composition_type(O=1, H=2), composition_type("OH2"))
            self.assertEqual(composition_type("(N)(C[12]H3)2(H)"), composition_type("NC[12]2H6H"))
            self.assertRaises(
                composition.ChemicalCompositionError, lambda: composition_type("O2.2"))

        def test_operators(self):
            case = composition_type("H2O")
            self.assertEqual(case + {'O': 1}, composition_type("H2O2"))
            self.assertEqual({'O': 1} + case, composition_type("H2O2"))

            self.assertEqual(case - {'O': 1}, composition_type("H2"))
            self.assertEqual({'O': 1, "H": 4} - case, composition_type("H2"))

            self.assertEqual(case * 3, composition_type("H6O3"))
            self.assertEqual(case * -1, -case)
            self.assertRaises(
                composition.ChemicalCompositionError, lambda: case * 5.2)

        def test_massing(self):
            case = composition_type("H2O")
            mono_mass = case.calc_mass()
            avg_mass = case.calc_mass(average=True)
            protonated = case + {"H+": 1}
            prot_mass = protonated.calc_mass()
            self.assertNotEqual(mono_mass, avg_mass)
            self.assertAlmostEqual(mono_mass, 18.0105, 3)
            self.assertAlmostEqual(avg_mass, 18.01528, 3)
            self.assertNotEqual(mono_mass, prot_mass)
            self.assertAlmostEqual(prot_mass, 19.01784, 3)
            self.assertAlmostEqual(case.calc_mass(charge=1), 19.01784, 3)
            self.assertRaises(
                composition.ChemicalCompositionError, lambda: protonated.calc_mass(charge=1))

        def test_pickling(self):
            case = composition_type("H2O")
            self.assertEqual(case, pickle.loads(pickle.dumps(case)))

        def test_most_probable_isotopic_composition(self):
            case = composition_type("H2O")
            comp, probability = composition.most_probable_isotopic_composition(case)
            self.assertEqual(comp, {'H[1]': 2, 'O[16]': 1})
            self.assertAlmostEqual(probability, 0.997, 3)

    return CompositionTests


from glypy.composition.composition import PComposition
PCompositionTests = make_composition_suite(PComposition)
try:
    from glypy.composition.composition import CComposition
    CCompositionTests = make_composition_suite(CComposition)
except ImportError:
    pass

if __name__ == '__main__':
    unittest.main()



import re
from collections import namedtuple

from glypy.composition import molecular_graph, Composition, formula
from glypy.structure import Substituent


def tokenize_alin(code):
    tokens = []
    current_token = ''
    i = 0
    n = len(code)
    connector_symbols = ('=', '/', '#', '$')
    while i < n:
        c = code[i]
        is_asterisk = c == '*'
        if c.isalpha() or is_asterisk:
            if c.isupper() or is_asterisk:
                if current_token:
                    tokens.append(current_token)
                    current_token = ''
                current_token += c
        elif c in ('^', '~'):
            current_token += code[i:i + 2]
            i += 2
            continue
        elif c in connector_symbols:
            if current_token:
                tokens.append(current_token)
                current_token = ''
            current_token += c
        elif c.isdigit():
            current_token += c

        i += 1
    if current_token:
        tokens.append(current_token)
    return tokens


def parse_alin(code):
    graph = molecular_graph.MolecularGraph()
    last_vertex = None
    bond_type = 1

    node_id_counter = 0
    for i, token in enumerate(tokenize_alin(code), 1):
        if token[0] in ('/', '$'):
            last_vertex = graph.vertices[int(token[1:])]
        elif token == '=':
            bond_type = 2
        elif token == '#':
            bond_type = 3
        else:
            properties = {}
            if "^" in token:
                isomorphisms = re.findall(r'[\^~][XSREZ]', token)
                token = re.sub(r'[\^~][XSREZ]', '', token)
                properties['stereoisomerism'] = isomorphisms
            if "*" in token:
                properties["is_backbone_carbon"] = True
                token = 'C'
            node_id_counter += 1
            vertex = molecular_graph.Atom(token, node_id_counter, **properties)
            graph.add_vertex(vertex)
            if last_vertex is not None:
                graph.add_edge(molecular_graph.Bond([last_vertex.id, vertex.id], bond_type))
            bond_type = 1
            last_vertex = vertex
    return graph


def alin_to_substituent(alin):
    if "*" in alin:
        name_query = alin.replace("*", "")
    else:
        name_query = alin
    try:
        record = map_to_substituent[name_query]
        name = record.name
    except KeyError:
        raise KeyError("Could not locate name for substituent described by %r" % (alin,))
    graph = parse_alin(alin)
    composition = Composition()
    unpaired_electrons = 2 if alin.startswith("*") else 1
    for node in graph:
        if node.data.get("is_backbone_carbon"):
            unpaired_electrons -= 1
            continue
        composition[node.element.symbol] += 1
        unpaired_electrons += node.unpaired_electrons
    composition["H"] += unpaired_electrons
    return Substituent(name, composition=composition)


def substituent_to_alin(substituent):
    try:
        name = substituent.name
    except AttributeError:
        name = str(substituent)
    record = substituent_to_map[name]
    return record.map_for_bmu


_SubstituentTranslation = namedtuple(
    "SubstituentTranslation", ("name", "map_for_bmu", "is_carbon_swap"))


class SubstituentTranslation(_SubstituentTranslation):
    @property
    def formula(self):
        graph = parse_alin(self.map_for_bmu)
        return formula(graph.total_composition())

    @property
    def internal_name(self):
        subst = Substituent.internalize_name(self.name)
        return subst


_substituent_translation = [
    SubstituentTranslation('acetyl', 'CC/2=O', None),
    SubstituentTranslation('amino', 'N', None),
    SubstituentTranslation('anhydro', '', None),
    SubstituentTranslation('bromo', 'Br', None),
    SubstituentTranslation('chloro', 'Cl', None),
    SubstituentTranslation('epoxy', '', None),
    SubstituentTranslation('ethanolamine', 'NCCO', False),
    SubstituentTranslation('ethyl', 'CC', None),
    SubstituentTranslation('fluoro', 'F', None),
    SubstituentTranslation('formyl', 'C=O', None),
    SubstituentTranslation('glycolyl', 'CCO/2=O', None),
    SubstituentTranslation('hydroxymethyl', 'CO', None),
    SubstituentTranslation('imino', '=N', False),
    SubstituentTranslation('iodo', 'I', None),
    SubstituentTranslation('lactone', '', None),
    SubstituentTranslation('methyl', 'C', None),
    SubstituentTranslation('n-acetyl', 'NCC/3=O', None),
    SubstituentTranslation('n-alanine', 'NCC^XC/4N/3=O', None),
    SubstituentTranslation('n-dimethyl', 'NC/2C', None),
    SubstituentTranslation('n-formyl', 'NC=O', None),
    SubstituentTranslation('n-glycolyl', 'NCCO/3=O', None),
    SubstituentTranslation('n-methyl', 'NC', None),
    SubstituentTranslation('n-succinate', 'NCCCCO/6=O/3=O', None),
    SubstituentTranslation('succinate', 'CCCCO/5=O/2=O', None),
    SubstituentTranslation('n-sulfate', 'NSO/3=O/3=O', True),
    SubstituentTranslation('n-triflouroacetyl', 'NCCF/4F/4F/3=O', None),
    SubstituentTranslation('nitrate', 'C=O/2=O', None),
    SubstituentTranslation('phosphate', 'OPO/3O/3=O', None),
    SubstituentTranslation('pyruvate', None, None),
    SubstituentTranslation('pyrophosphate', 'P^XOPO/4O/4=O/2O/2=O', None),
    SubstituentTranslation('triphosphate', 'P^XOP^XOPO/6O/6=O/4O/4=O/2O/2=O', None),
    SubstituentTranslation('(r)-lactate', 'CC^RC/3O/2=O', None),
    SubstituentTranslation('(r)-pyruvate', None, None),
    SubstituentTranslation('(s)-lactate', 'CC^SC/3O/2=O', None),
    SubstituentTranslation('(s)-pyruvate', None, None),
    SubstituentTranslation('sulfate', 'OSO/3=O/3=O', None),
    SubstituentTranslation('thio', 'S', None),
    SubstituentTranslation('amidino', 'CN/2=N', None),
    SubstituentTranslation('n-amidino', 'NCN/3=N', None),
    SubstituentTranslation('(r)-carboxymethyl', '?*', None),
    SubstituentTranslation('carboxymethyl', 'CCO/3=O', None),
    SubstituentTranslation('(s)-carboxymethyl', '?*', None),
    SubstituentTranslation('(r)-carboxyethyl', 'C^RCO/3=O/2C', None),
    SubstituentTranslation('(s)-carboxyethyl', 'C^SCO/3=O/2C', None),
    SubstituentTranslation('n-methyl-carbamoyl', 'CNC/2=O', None),
    SubstituentTranslation('phospho-ethanolamine', 'P^XOCCN/2O/2=O', True),
    SubstituentTranslation('phospho-ethanolamine', 'OP^XOCCN/3O/3=O', True),
    SubstituentTranslation('diphospho-ethanolamine', 'P^XOP^XOCCN/4O/4=O/2O/2=O', True),
    SubstituentTranslation('phospho-choline', 'P^XOCCN/5N/5N/2O/2=O', None),
    SubstituentTranslation('(x)-lactate', 'CC^XC/3O/2=O', None),
    SubstituentTranslation('(r)-1-hydroxymethyl', '?*', None),
    SubstituentTranslation('(s)-1-hydroxymethyl', '?*', None),
]


substituent_to_map = {
    unit.internal_name: unit for unit in _substituent_translation
}


map_to_substituent = {
    unit.map_for_bmu: unit for unit in _substituent_translation
    if unit.map_for_bmu is not None
}


formula_to_substituent = {
    unit.formula: unit for unit in _substituent_translation
    if unit.map_for_bmu
}



# pragma: no cover
import gzip
import logging
import requests
from lxml import etree

from glypy.utils import StringIO
from glypy.io import glycoct
from glypy.algorithms.database import (Taxon, Aglyca, Motif,
                                       DatabaseEntry, GlycanRecord,
                                       GlycanRecordWithTaxon,
                                       RecordDatabase)

try:
    xrange
except NameError:
    xrange = range


import warnings
warnings.warn("Glycome-DB is no longer available. All Methods Will Fail.", DeprecationWarning)


logger = logging.getLogger(__name__)

cache = None


def download_all_structures(db_path, record_type=GlycanRecordWithTaxon):  # pragma: no cover
    response = requests.get(u'http://www.glycome-db.org/http-services/getStructureDump.action?user=eurocarbdb')
    response.raise_for_status()
    handle = gzip.GzipFile(fileobj=StringIO(response.content))
    xml = etree.parse(handle)
    db = RecordDatabase(db_path, record_type=record_type)
    misses = []
    i = 0
    for structure in xml.iterfind(".//structure"):
        try:
            glycomedb_id = int(structure.attrib['id'])
            i += 1
            glycoct_str = structure.find("sequence").text
            taxa = [Taxon(t.attrib['ncbi'], None, None) for t in structure.iterfind(".//taxon")]
            glycan = glycoct.loads(glycoct_str)
            if (glycoct.loads(str(glycan)).mass() - glycan.mass()) > 0.00001:
                raise Exception("Mass did not match on reparse")
            record = record_type(glycan, taxa=taxa, id=glycomedb_id)
            db.load_data(record, commit=False, set_id=False)
            if i % 1000 == 0:
                print(i, "Records parsed.")
        except Exception as e:
            misses.append((glycomedb_id, e))
            print(glycomedb_id, e)
    db.set_metadata("misses", misses)
    db.commit()
    return db


def set_cache(path):
    '''Set the path to the current cache for use with this client.
    This will create an instance of :class:`RecordDatabase` using the
    path provided.

    Any record lookups will first check the cache for that identifier, and
    if found will return from there. Any records not found will be fetched from
    the remote database and saved in the cache before returning.
    '''
    global cache
    logger.info("Setting glycomedb client cache to %r", path)
    cache = RecordDatabase(path)


def add_cache(record):
    if cache is None:
        return
    for stmt in record.to_sql():
        try:
            cache.connection.execute(stmt)
        except Exception as e:
            logger.error("An error occurred while adding %r", record, exc_info=e)


def check_cache(key):
    if cache is None:
        return False
    res = int(key) in cache
    if not res:
        logger.debug("Cache miss %r", key)
    else:
        logger.debug("Cache hit %r", key)
    return res


get_url_template = "http://www.glycome-db.org/database/showStructure.action?glycomeId={id}"
xpath = ".//condenced"


def _get_raw(id):
    r = requests.get(get_url_template.format(id=id))
    r.raise_for_status()
    tree = etree.fromstring(r.content)
    condensed = tree.find(xpath).text
    return condensed


def get(id):
    '''
    Get the structure for `id` from :title-reference:`GlycomeDB`.

    GlycomeDB supplies a detailed schema link which allows `lxml` to easily pull out
    more than just the GlycoCT string. To download a more informative record, use :func:`get_record`

    Parameters
    ----------
    id: str or int

    Returns
    -------
    Glycan
    '''
    if check_cache(id):
        return cache[id].structure
    r = requests.get(get_url_template.format(id=id))
    r.raise_for_status()
    tree = etree.fromstring(r.content)
    condensed = tree.find(xpath).text
    return glycoct.loads(condensed)


#: GlycomeDB supplies a detailed schema link which allows `lxml` to easily pull out
#: more than just the GlycoCT string. To download a more informative record, use :func:`get_record`

def get_record(id):
    '''
    Get the complete record for `id` from :title-reference:`GlycomeDB`.

    Returns
    -------
    GlycanRecord
    '''
    if check_cache(id):
        return cache[id]
    r = requests.get(get_url_template.format(id=id))
    r.raise_for_status()
    tree = etree.fromstring(r.content)
    return glycan_record_from_xml(tree, id)


def search_substructure(glycan_obj):  # pragma: no cover
    '''
    Select all structures which have contain `glycan_obj`.

    Parameters
    ----------
    glycan_obj: Glycan
        The glycan structure to search for

    Yields
    ------
    GlycomeDBSearchMatch
    '''
    post_data = {
        "sequencetype": "glycoct_condenced",
        "sequence": glycan_obj.to_glycoct(),
        "msView": "yes"
    }
    session = requests.Session()
    r = session.post(_search_url, data=post_data)
    r.raise_for_status()
    tree = etree.fromstring(r.content)
    try:
        last_page = int([p.attrib for p in tree.findall('.//page')][-1]["number"])
    except:
        last_page = 1
    matches = [GlycomeDBSearchMatch(**s.attrib) for s in tree.findall(".//structure")]
    for m in matches:
        yield m
    for page_id in range(2, last_page):
        get_more = session.post(_get_more_results_url, data={"page": page_id})
        get_more.raise_for_status()
        tree = etree.fromstring(get_more.content)
        matches = [GlycomeDBSearchMatch(**s.attrib) for s in tree.findall(".//structure")]
        if len(matches) == 0:
            raise StopIteration(get_more.content)
        for m in matches:
            yield m


_search_url = "http://www.glycome-db.org/database/searchSubStructure.action"
_get_more_results_url = "http://www.glycome-db.org/database/showMultiStructure.action"


def search_minimum_common_substructure(glycan_obj, minimum_residues=1):  # pragma: no cover
    '''
    Select all structures which have a minimum common substructure of `minimum_residues`
    with `glycan_obj`.

    Parameters
    ----------
    glycan_obj: Glycan
        The glycan structure to search for
    minimum_residues: int
        The minimum number of residues to match

    Yields
    ------
    GlycomeDBSearchMatch
    '''
    post_data = {
        "sequencetype": "glycoct_condenced",
        "sequence": glycan_obj.to_glycoct(),
        "msView": "yes",
        "minResidueCount": minimum_residues
    }
    session = requests.Session()
    r = session.post(_mcs_search_url, data=post_data)
    r.raise_for_status()
    tree = etree.fromstring(r.content)
    try:
        last_page = int([p.attrib for p in tree.findall('.//page')][-1]["number"])
    except:
        last_page = 1
    matches = [GlycomeDBSearchMatch(**s.attrib) for s in tree.findall(".//structure")]
    for m in matches:
        yield m
    for page_id in xrange(2, last_page):
        get_more = session.post(_get_more_results_url, data={"page": page_id})
        get_more.raise_for_status()
        tree = etree.fromstring(get_more.content)
        matches = [GlycomeDBSearchMatch(**s.attrib) for s in tree.findall(".//structure")]
        if len(matches) == 0:
            raise StopIteration(get_more.content)
        for m in matches:
            yield m


_mcs_search_url = "http://www.glycome-db.org/database/searchMCS.action"


def search_by_species(tax_id):
    '''
    Select all structures which are associated with the provided taxonomy id, or
    one of its children.

    Parameters
    ----------
    tax_id: int or str
        The taxonomy id to search for

    Yields
    ------
    GlycomeDBSearchMatch
    '''
    post_data = {
        "species": tax_id,
        "type": "ncbi",
        "subTree": "yes",
    }
    session = requests.Session()
    r = session.post(_taxa_search_url, data=post_data)
    r.raise_for_status()
    tree = etree.fromstring(r.content)
    try:
        last_page = int([p.attrib for p in tree.findall('.//page')][-1]["number"])
    except:
        last_page = 1
    matches = [GlycomeDBSearchMatch(**s.attrib) for s in tree.findall(".//structure")]
    for m in matches:
        yield m
    for page_id in xrange(2, last_page):
        get_more = session.post(_get_more_results_url, data={"page": page_id})
        get_more.raise_for_status()
        tree = etree.fromstring(get_more.content)
        matches = [GlycomeDBSearchMatch(**s.attrib) for s in tree.findall(".//structure")]
        if len(matches) == 0:
            session.close()
            # raise StopIteration(get_more.content)
            return
        for m in matches:
            yield m
    session.close()


_taxa_search_url = "http://www.glycome-db.org/database/searchBySpecies.action"


def download_results(results, to=None):  # pragma: no cover
    if to is not None:
        if isinstance(to, str):
            set_cache(to)
        elif isinstance(to, RecordDatabase):
            global cache
            cache = to
    for r in results:
        print(r.id)
        try:
            r.get_record()
        except glycoct.GlycoCTError as e:
            logger.exception("Exception %r", e, exc_info=e)


class GlycomeDBSearchMatch(object):
    '''
    A search match which carries information about the scored match and the
    id number for retrieving the |Glycan| or |GlycanRecord|.
    '''
    def __init__(self, id, score, cross_reference, species_number, **kwargs):
        self.id = int(id)
        self.score = score
        self.num_crossreferences = cross_reference
        self.num_species = species_number

    def get(self):
        """Fetch the |Glycan| referenced by this search result

        Returns
        -------
        Glycan:
            The |Glycan| referenced
        """
        return get(self.id)

    def get_record(self):
        """Fetch the |GlycanRecord| referenced by this search result

        Returns
        -------
        GlycanRecord:
            The |GlycanRecord| referenced by this search result
        """
        return get_record(self.id)


def make_entries(annotation):
    '''
    Helper method to create :class:`Entry` objects from <entry></entry> tags in :class:`Taxon` and :class:`Aglyca`.
    '''
    res = [DatabaseEntry(node.attrib['database'], node.attrib['id']) for node in annotation.findall(".//entry")]
    return res


def glycan_record_from_xml(xml_tree, id):
    '''
    Converts an XML document and the associated database into an instance of
    `GlycanRecord`.

    Parameters
    ----------
    xml_tree: lxml.etree
        XML document to consume
    id:
        GlycomeDB id number to assign this record

    Returns
    -------
    GlycanRecord:
        Constructed record
    '''
    structure = glycoct.loads(xml_tree.find(xpath).text)
    taxa = [Taxon(t.attrib['ncbi'], t.attrib['name'], make_entries(t)) for t in xml_tree.findall(".//taxon")]
    aglycon = [Aglyca(t.attrib['name'].replace(
        "'", "`"), t.attrib['reducing'], make_entries(t)) for t in xml_tree.findall(".//aglyca")]
    motifs = [Motif(t.attrib['name'], t.attrib['id'], t.attrib['class']) for t in xml_tree.findall(".//motif")]
    dbxref = [e for c in [t.entries for t in taxa] + [t.entries for t in aglycon] for e in c]
    dbxref.append(DatabaseEntry("GlycomeDB", id))
    record = GlycanRecord(structure, motifs=motifs, dbxref=dbxref, aglycones=aglycon, taxa=taxa, id=id)
    record.id = id
    add_cache(record)
    return record

if __name__ == "__main__":
    import sys
    download_all_structures(sys.argv[1])



import sys
cimport cython
from cpython cimport PyErr_SetString, PyObject_Hash
from cpython.mem cimport PyMem_Malloc, PyMem_Free, PyMem_Realloc
from cpython.ref cimport PyObject, Py_INCREF, Py_DECREF, Py_XDECREF, Py_XINCREF
from cpython.int cimport PyInt_AsLong, PyInt_FromLong
from cpython.dict cimport PyDict_Next, PyDict_SetItem

try:
    from collections import Mapping, MutableMapping
except ImportError:
    from collections.abc import Mapping, MutableMapping


cdef int initialize_count_table_bin(count_table_bin* bin, size_t size):
    bin.cells = <count_table_bin_cell*>PyMem_Malloc(sizeof(count_table_bin_cell) * size)
    if bin.cells == NULL:
        return 1
    for i in range(size):
        bin.cells[i].key = NULL
    bin.used = 0
    bin.size = size


cdef void free_count_table_bin(count_table_bin* bin):
    for i in range(bin.used):
        if bin.cells[i].key != NULL:
            Py_XDECREF(bin.cells[i].key)
            bin.cells[i].key = NULL
    PyMem_Free(bin.cells)


cdef int count_table_bin_append(count_table_bin* bin, PyObject* key, long value):
    if bin.used == bin.size - 1:
        bin.cells = <count_table_bin_cell*>PyMem_Realloc(bin.cells, sizeof(count_table_bin_cell) * bin.size * 2)
        if bin.cells == NULL:
            return 1
        bin.size *= 2
    Py_XINCREF(key)
    bin.cells[bin.used].key = key
    bin.cells[bin.used].value = value
    bin.used += 1
    return 0


cdef int count_table_bin_find(count_table_bin* bin, PyObject* query, Py_ssize_t* cell_index):
    cdef:
        object query_obj
    query_obj = <object>query
    for i in range(bin.used):
        if bin.cells[i].key == NULL:
            continue
        Py_XINCREF(bin.cells[i].key)
        if (bin.cells[i].key == (<PyObject*>query_obj)) or ((<object>bin.cells[i].key) == (query_obj)):
            Py_XDECREF(bin.cells[i].key)
            cell_index[0] = i
            return 0
        Py_XDECREF(bin.cells[i].key)
    cell_index[0] = -1
    return 0


cdef count_table* make_count_table(size_t table_size, size_t bin_size):
    cdef:
        size_t i, j
        count_table* table
    table = <count_table*>PyMem_Malloc(sizeof(count_table))
    if table == NULL:
        PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
        return NULL
    table.bins = <count_table_bin*>PyMem_Malloc(sizeof(count_table_bin) * table_size)
    if table.bins == NULL:
        PyErr_SetString(MemoryError, "Could not allocate memory for count_table")
        return NULL
    for i in range(table_size):
        initialize_count_table_bin(&(table.bins[i]), bin_size)
    table.size = table_size
    return table


cdef void free_count_table(count_table* table):
    for i in range(table.size):
        free_count_table_bin(&table.bins[i])
    PyMem_Free(table.bins)
    PyMem_Free(table)


@cython.cdivision(True)
cdef int count_table_find_bin(count_table* table, PyObject* query, Py_ssize_t* bin_index) except 1:
    try:
        bin_index[0] = PyObject_Hash(<object>query) % table.size
    except TypeError:
        PyErr_SetString(TypeError, "%r is not hashable" % (<object>query, ))
        return 1
    return 0


cdef int count_table_put(count_table* table, PyObject* key, long value) except 1:
    cdef:
        Py_ssize_t bin_index, cell_index
        int status
    if key == NULL:
        return 1
    Py_XINCREF(key)
    status = count_table_find_bin(table, key, &bin_index)
    if status != 0:
        return 1
    status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
    if cell_index == -1:
        status = count_table_bin_append(&table.bins[bin_index], key, value)
        Py_XDECREF(key)
        if status != 0:
            return 1
    else:
        table.bins[bin_index].cells[cell_index].value = value
        Py_XDECREF(key)
    return 0


cdef int count_table_del(count_table* table, PyObject* key, long* value) except 1:
    cdef:
        Py_ssize_t bin_index, cell_index
        int status
    Py_XINCREF(key)
    value[0] = 0
    status = count_table_find_bin(table, key, &bin_index)
    if status != 0:
        return 1
    status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
    if cell_index == -1:
        value[0] = 0
        Py_XDECREF(key)
        return 0
    else:
        value[0] = table.bins[bin_index].cells[cell_index].value
        Py_XDECREF(table.bins[bin_index].cells[cell_index].key)
        table.bins[bin_index].cells[cell_index].key = NULL
        table.bins[bin_index].cells[cell_index].value = 0
        Py_XDECREF(key)
        return 0


cdef int count_table_get(count_table* table, PyObject* key, long* value) except 1:
    cdef:
        Py_ssize_t bin_index, cell_index
        int status
    value[0] = 0
    Py_XINCREF(key)
    status = count_table_find_bin(table, key, &bin_index)
    if status != 0:
        return 1
    status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
    Py_XDECREF(key)
    if cell_index == -1:
        value[0] = 0
        return 0
    else:
        value[0] = table.bins[bin_index].cells[cell_index].value
        return 0


cdef int count_table_increment(count_table* table, PyObject* key, long value):
    cdef:
        Py_ssize_t bin_index, cell_index
        int status
    Py_XINCREF(key)
    status = count_table_find_bin(table, key, &bin_index)
    status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
    if cell_index == -1:
        status = count_table_bin_append(&table.bins[bin_index], key, value)
        Py_XDECREF(key)
        if status != 0:
            return 1
    else:
        table.bins[bin_index].cells[cell_index].value += value
        Py_XDECREF(key)
    return 0


cdef int count_table_decrement(count_table* table, PyObject* key, long value):
    cdef:
        Py_ssize_t bin_index, cell_index
        int status
    Py_XINCREF(key)
    status = count_table_find_bin(table, key, &bin_index)
    status = count_table_bin_find(&table.bins[bin_index], key, &cell_index)
    if cell_index == -1:
        status = count_table_bin_append(&table.bins[bin_index], key, -value)
        Py_XDECREF(key)
        if status != 0:
            return 1
    else:
        table.bins[bin_index].cells[cell_index].value -= value
        Py_XDECREF(key)
    return 0


cdef Py_ssize_t count_table_count(count_table* table):
    cdef:
        Py_ssize_t count
        size_t i, j
    count = 0
    for i in range(table.size):
        for j in range(table.bins[i].used):
            count += table.bins[i].cells[j].key != NULL
    return count


cdef list count_table_keys(count_table* table):
    cdef:
        list keys
        size_t i, j
    keys = []
    count = 0
    for i in range(table.size):
        for j in range(table.bins[i].used):
            if table.bins[i].cells[j].key != NULL:
                keys.append(<object>table.bins[i].cells[j].key)
    return keys


cdef list count_table_values(count_table* table):
    cdef:
        list values
        size_t i, j
    values = []
    count = 0
    for i in range(table.size):
        for j in range(table.bins[i].used):
            if table.bins[i].cells[j].key != NULL:
                values.append(PyInt_FromLong(table.bins[i].cells[j].value))
    return values


cdef list count_table_items(count_table* table):
    cdef:
        list items
        size_t i, j
    items = []
    for i in range(table.size):
        for j in range(table.bins[i].used):
            if table.bins[i].cells[j].key != NULL:
                items.append(
                    (<object>table.bins[i].cells[j].key,
                     PyInt_FromLong(table.bins[i].cells[j].value)))
    return items


cdef void count_table_add(count_table* table_a, count_table* table_b):
    cdef:
        size_t i, j
        long value
    for i in range(table_b.size):
        for j in range(table_b.bins[i].used):
            if table_b.bins[i].cells[j].key != NULL:
                count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
                count_table_increment(table_a, table_b.bins[i].cells[j].key, value)


cdef void count_table_subtract(count_table* table_a, count_table* table_b):
    cdef:
        size_t i, j
        long value, temp
    for i in range(table_b.size):
        for j in range(table_b.bins[i].used):
            if table_b.bins[i].cells[j].key != NULL:
                count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
                count_table_decrement(table_a, table_b.bins[i].cells[j].key, value)


cdef void count_table_scale(count_table* table, long value):
    cdef:
        size_t i, j
    for i in range(table.size):
        for j in range(table.bins[i].used):
            if table.bins[i].cells[j].key != NULL:
                table.bins[i].cells[j].value *= value


cdef void count_table_update(count_table* table_a, count_table* table_b):
    cdef:
        size_t i, j
        long value
    for i in range(table_b.size):
        for j in range(table_b.bins[i].used):
            if table_b.bins[i].cells[j].key != NULL:
                count_table_get(table_b, table_b.bins[i].cells[j].key, &value)
                count_table_put(table_a, table_b.bins[i].cells[j].key, value)


cdef void count_table_clear(count_table* table):
    cdef:
        size_t i, j
    for i in range(table.size):
        for j in range(table.bins[i].used):
            Py_XDECREF(table.bins[i].cells[j].key)
            table.bins[i].cells[j].key = NULL
            table.bins[i].cells[j].value = 0


cdef count_table* count_table_copy(count_table* table_a):
    cdef:
        count_table* dup
        size_t i, j
    dup = make_count_table(table_a.size, 2)
    for i in range(table_a.size):
        for j in range(table_a.bins[i].used):
            if table_a.bins[i].cells[j].key != NULL:
                count_table_bin_append(
                    &dup.bins[i], table_a.bins[i].cells[j].key,
                    table_a.bins[i].cells[j].value)
    return dup


cdef dict count_table_to_dict(count_table* table_a):
    cdef:
        dict dup
        size_t i, j
    dup = dict()
    for i in range(table_a.size):
        for j in range(table_a.bins[i].used):
            if table_a.bins[i].cells[j].key != NULL:
                PyDict_SetItem(
                    dup, <object>table_a.bins[i].cells[j].key,
                    PyInt_FromLong(table_a.bins[i].cells[j].value))
    return dup


cdef bint count_table_equals(count_table* table_a, count_table* table_b):
    cdef:
        size_t i, j
        long value
    for i in range(table_a.size):
        for j in range(table_a.bins[i].used):
            if table_a.bins[i].cells[j].key != NULL:
                count_table_get(table_b, table_a.bins[i].cells[j].key, &value)
                if table_a.bins[i].cells[j].value != value:
                    return False
    for i in range(table_b.size):
        for j in range(table_b.bins[i].used):
            if table_b.bins[i].cells[j].key != NULL:
                count_table_get(table_a, table_b.bins[i].cells[j].key, &value)
                if table_b.bins[i].cells[j].value != value:
                    return False
    return True


cdef class CountTableIterator(object):

    def __init__(self, table):
        self.table = (<CountTable>table).table
        self.bin_index = 0
        self.cell_index = 0
        self.next_key = NULL
        self.next_value = 0
        self.advance()

    @staticmethod
    cdef CountTableIterator _create(CountTable table):
        cdef CountTableIterator it = CountTableIterator.__new__(CountTableIterator)
        it.table = table.table
        it.bin_index = 0
        it.cell_index = 0
        it.next_key = NULL
        it.next_value = 0
        it.advance()
        return it

    cdef bint has_more(self):
        return self.next_key != NULL

    cdef int _locate_next_value(self, PyObject** key, long* value):
        if not self.bin_index < self.table.size:
            key[0] = NULL
            return 1
        # start from the previous iteration's bin
        for i in range(self.cell_index, self.table.bins[self.bin_index].used):
            self.cell_index = i
            if self.table.bins[self.bin_index].cells[i].key != NULL:
                key[0] = self.table.bins[self.bin_index].cells[i].key
                value[0] = self.table.bins[self.bin_index].cells[i].value
                self.cell_index += 1
                return 0
        # move onto the next bins
        for i in range(self.bin_index + 1, self.table.size):
            self.bin_index = i
            for j in range(0, self.table.bins[i].used):
                self.cell_index = j
                if self.table.bins[i].cells[j].key != NULL:
                    key[0] = self.table.bins[self.bin_index].cells[j].key
                    value[0] = self.table.bins[self.bin_index].cells[j].value
                    self.cell_index += 1
                    return 0
        self.bin_index = self.table.size
        key[0] = NULL
        return 1

    cdef int get_next_value(self, PyObject** key, long* value):
        cdef:
            PyObject* next_key
            long next_value
            int status
        if self.next_key == NULL:
            key[0] = NULL
            value[0] = 0
            return 1
        else:
            key[0] = self.next_key
            value[0] = self.next_value
            status = self._locate_next_value(&next_key, &next_value)
            self.next_key = next_key
            self.next_value = next_value
        return 0

    cdef void advance(self):
        cdef:
            PyObject* next_key
            long next_value
            int status
        status = self._locate_next_value(&next_key, &next_value)
        self.next_key = next_key
        self.next_value = next_value

    def __iter__(self):
        return self

    def __next__(self):
        cdef:
            PyObject* key
            PyObject* next_key
            long value, next_value
            int status
        if self.next_key == NULL:
            raise StopIteration()
        else:
            key = self.next_key
            value = self.next_value
            status = self._locate_next_value(&next_key, &next_value)
            self.next_key = next_key
            self.next_value = next_value
        return (<object>key, value)

    def _test(self):
        cdef:
            PyObject* key
            long count
            int pos, j
            list result
        pos = 0
        j = 0
        result = []
        while self.has_more():
            pos = self.get_next_value(&key, &count)
            if pos != 0:
                break
            result.append((<object>key, count))
            j += 1
        return result



cdef class CountTable(object):

    @staticmethod
    cdef CountTable _create():
        cdef CountTable inst
        inst = CountTable.__new__(CountTable)
        inst._initialize_table()
        return inst

    @staticmethod
    cdef CountTable _create_from(CountTable table):
        cdef CountTable inst
        inst = CountTable.__new__(CountTable)
        inst.table = count_table_copy(table.table)
        return inst        

    def __init__(self, obj=None, **kwargs):
        self._initialize_table()
        if obj is not None:
            self.update(obj)
        if kwargs:
            self.update(kwargs)

    cdef int _initialize_table(self) except 1:
        self.table = make_count_table(6, 2)
        if self.table == NULL:
            return 1
        return 0

    cpdef update(self, obj):
        if isinstance(obj, CountTable):
            self._update_from_count_table(<CountTable>obj)
        if isinstance(obj, dict):
            self._update_from_dict(<dict>obj)
        else:
            for k, v in obj.items():
                self[k] = v

    cdef void _update_from_dict(self, dict d):
        cdef:
            PyObject *key
            PyObject *value
            Py_ssize_t pos
        pos = 0
        while PyDict_Next(d, &pos, &key, &value):
            self.setitem(<object>key, PyInt_AsLong(int(<object>value)))

    cdef void _update_from_count_table(self, CountTable other):
        count_table_update(self.table, other.table)

    cpdef CountTable copy(self):
        cdef CountTable inst = CountTable._create_from(self)
        return inst

    def __reduce__(self):
        return self.__class__, (self._to_dict(), )

    def __dealloc__(self):
        free_count_table(self.table)

    def __getitem__(self, object key):
        cdef long value = self.getitem(key)
        return PyInt_FromLong(value)

    def __setitem__(self, object key, object value):
        self.setitem(key, PyInt_AsLong(int(value)))

    def __delitem__(self, object key):
        self.pop(key)

    def __len__(self):
        return count_table_count(self.table)

    def __contains__(self, key):
        cdef long value = self.getitem(key)
        return value != 0

    def __iter__(self):
        return iter(self.keys())

    def __add__(self, other):
        cdef:
            CountTable dup
        if isinstance(other, CountTable):
            if isinstance(self, CountTable):
                dup = (<CountTable>self).copy()
                dup._add_from(<CountTable>other)
                return dup
            elif isinstance(self, dict):
                dup = (<CountTable>other).copy()
                dup._add_from_dict(<dict>self)
                return dup
            else:
                self = CountTable(self)
                dup = (<CountTable>other).copy()
                dup._add_from(self)
                return dup
        else:
            if isinstance(other, dict):
                dup = (<CountTable>self).copy()
                dup._add_from_dict(<dict>other)
                return dup
            else:
                dup = (<CountTable>self).copy()
                dup._add_from(CountTable(other))
                return dup

    def __iadd__(self, other):
        if isinstance(other, CountTable):
            self._add_from(<CountTable>other)
        elif isinstance(other, dict):
            self._add_from_dict(<dict>other)
        else:
            self._add_from(CountTable(other))
        return self

    def __sub__(self, other):
        cdef:
            CountTable dup
        if isinstance(other, CountTable):
            if isinstance(self, CountTable):
                dup = (<CountTable>self).copy()
                dup._subtract_from(<CountTable>other)
                return dup
            elif isinstance(self, dict):
                dup = (<CountTable>other).copy()
                dup._subtract_from_dict(<dict>self)
                return dup
            else:
                self = CountTable(self)
                dup = (<CountTable>other).copy()
                dup._subtract_from(self)
                return dup
        else:
            if isinstance(other, dict):
                dup = (<CountTable>self).copy()
                dup._subtract_from_dict(<dict>other)
                return dup
            else:
                dup = (<CountTable>self).copy()
                dup._subtract_from(CountTable(other))
                return dup

    def __isub__(self, other):
        if isinstance(other, CountTable):
            self._subtract_from(<CountTable>other)
        elif isinstance(other, dict):
            self._subtract_from_dict(<dict>other)
        else:
            self._subtract_from(CountTable(other))
        return self

    def __mul__(self, other):
        cdef:
            CountTable dup
        if isinstance(other, CountTable):
            dup = (<CountTable>other).copy()
            dup._scale_by(self)
            return dup
        elif isinstance(self, CountTable):
            dup = (<CountTable>self).copy()
            dup._scale_by(other)
            return dup
        else:
            return NotImplemented

    def __imul__(self, other):
        self._scale_by(other)
        return self

    def __neg__(self):
        return self._scale_by(-1)

    cdef void _add_from(self, CountTable other):
        count_table_add(self.table, other.table)

    cdef int _add_from_dict(self, dict other) except 1:
        cdef:
            PyObject *key
            PyObject *value
            Py_ssize_t pos
            long v
        pos = 0
        try:
            while PyDict_Next(other, &pos, &key, &value):
                v = PyInt_AsLong(int(<object>value))
                self.setitem(
                    <object>key,
                    v + self.getitem(<object>key))
        except TypeError:
            PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
            return 1
        return 0

    cdef void _subtract_from(self, CountTable other):
        count_table_subtract(self.table, other.table)

    cdef int _subtract_from_dict(self, dict other) except 1:
        cdef:
            PyObject *key
            PyObject *value
            Py_ssize_t pos
            long v
        pos = 0
        try:
            while PyDict_Next(other, &pos, &key, &value):
                v = PyInt_AsLong(int(<object>value))
                self.setitem(
                    <object>key,
                    v - self.getitem(<object>key))
        except TypeError:
            PyErr_SetString(TypeError, "%r must be a number" % (<object>value, ))
            return 1
        return 0

    cdef void _scale_by(self, long value):
        count_table_scale(self.table, value)

    cpdef dict _to_dict(self):
        return count_table_to_dict(self.table)

    def __repr__(self):
        return "{self.__class__.__name__}({content})".format(
            self=self, content=self._to_dict())

    cdef bint equal_to(self, CountTable other):
        return count_table_equals(self.table, other.table)

    def __eq__(self, other):
        if isinstance(self, CountTable):
            if isinstance(other, CountTable):
                return count_table_equals(self.table, (<CountTable>other).table)
            else:
                return self._to_dict() == other
        else:
            return self == other._to_dict()

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        cdef:
            Py_hash_t hash_value
            size_t i, j

        hash_value = 0
        for i in range(self.table.size):
            for j in range(self.table.bins[i].used):
                if self.table.bins[i].cells[j].key != NULL:
                    hash_value ^= hash(<object>self.table.bins[i].cells[j].key)
                    hash_value ^= self.table.bins[i].cells[j].value
        hash_value ^= (hash_value >> 11) ^ (hash_value >> 25)
        hash_value = hash_value * 69069 + 907133923
        return hash_value

    cpdef list keys(self):
        return count_table_keys(self.table)

    cpdef list values(self):
        return count_table_values(self.table)

    cpdef list items(self):
        return count_table_items(self.table)

    cpdef clear(self):
        count_table_clear(self.table)

    cpdef setdefault(self, key, value):
        if self.getitem(key) == 0:
            self.setitem(key, value)

    cdef long getitem(self, object key) except *:
        cdef long value
        cdef int status 
        cdef PyObject* pkey = <PyObject*>key
        Py_INCREF(key)
        status = count_table_get(self.table, pkey, &value)
        Py_DECREF(key)
        if status != 0:
            return 0
        return value        

    cdef int setitem(self, object key, long value) except 1:
        cdef PyObject* pkey = <PyObject*>key
        Py_INCREF(key)
        cdef int status = count_table_put(self.table, pkey, value)
        Py_DECREF(key)
        if status != 0:
            return 1
        return 0

    cdef long delitem(self, object key) except *:
        cdef PyObject* pkey = <PyObject*>key
        cdef long value
        Py_INCREF(key)
        cdef int status = count_table_del(self.table, pkey, &value)
        Py_DECREF(key)
        if status != 0:
            return 0
        return value

    cdef void increment(self, object key, long value):
        cdef PyObject* pkey = <PyObject*>key
        Py_INCREF(key)
        cdef int status = count_table_increment(self.table, pkey, value)
        Py_DECREF(key)

    cdef void decrement(self, object key, long value):
        cdef PyObject* pkey = <PyObject*>key
        Py_INCREF(key)
        cdef int status = count_table_decrement(self.table, pkey, value)
        Py_DECREF(key)

    cpdef long pop(self, object key, object default=None) except *:
        cdef long value = self.delitem(key)
        if value == 0:
            return default
        return PyInt_FromLong(value)

    cpdef get(self, object key, object default=None):
        cdef long value = self.getitem(key)
        if value == 0:
            return default
        return PyInt_FromLong(value)


Mapping.register(CountTable)
MutableMapping.register(CountTable)


def main():
    cdef long val
    cdef str string
    cdef object pint
    cdef PyObject* key
    string = "spam"
    pint = 12
    key = <PyObject*>string
    cdef count_table* table = make_count_table(6, 2)
    count_table_put(table, key, 42)
    count_table_get(table, key, &val)
    assert val == 42
    key = <PyObject*>pint
    count_table_put(table, key, 67)
    count_table_del(table, key, &val)
    assert val == 67
    val = 1
    count_table_get(table, key, &val)
    free_count_table(table)
    pytable = CountTable()
    Py_INCREF(<object>key)
    pytable[<object>key] = 2
    pytable[int(12)] = 11
    pytable[255.2] = 99
    assert len(pytable) == 2
    del pytable
    pytable = CountTable._create()
    print pytable



import unittest
from glypy.io import glyspace
from glypy import tree, root

from pytest import mark

try:
    is_online = True
    response = glyspace.requests.get(glyspace.sparql_endpoint)
    response.raise_for_status()
except:
    is_online = False


def skip_not_online(fn):
    if not is_online:
        return mark.xfail(mark.skip("Not able to reach host")(fn))
    else:
        return mark.xfail(fn)


example_n3 = '''@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix glycan: <http://purl.jp/bio/12/glyco/glycan#> .
@prefix glycoinfo: <http://rdf.glycoinfo.org/glycan/> .
@prefix glycomedb: <http://rdf.glycome-db.org/glycan/> .
@prefix glytoucan: <http://www.glytoucan.org/glyco/owl/glytoucan#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .
@prefix xml: <http://www.w3.org/XML/1998/namespace> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

glycoinfo:G80903UK a glycan:saccharide ;
    glycan:has_glycosequence <http://rdf.glycoinfo.org/glycan/G80903UK/glycoct>,
        <http://rdf.glycoinfo.org/glycan/G80903UK/iupac_condensed>,
        <http://rdf.glycoinfo.org/glycan/G80903UK/iupac_extended>,
        <http://rdf.glycoinfo.org/glycan/G80903UK/wurcs/2.0> ;
    glycan:has_motif glycoinfo:G00031MO,
        glycoinfo:G00032MO,
        glycoinfo:G00033MO,
        glycoinfo:G00034MO,
        glycoinfo:G00041MO,
        glycoinfo:G00042MO ;
    glycan:has_resource_entry <http://rdf.glycoinfo.org/glycan/resource-entry/G80903UK/glyconnect/3641>,
        <http://rdf.glycoinfo.org/glycan/resource-entry/G80903UK/unicarbkb/10>,
        <http://rdf.glycoinfo.org/glycan/resource-entry/null/glytoucan/G80903UK> ;
    glytoucan:has_derivatized_mass <http://rdf.glycoinfo.org/derivatization_type_node/877.31754966> ;
    glytoucan:has_primary_id "G80903UK" .
'''


class LocalClient(glyspace.GlyTouCanRDFClient):
    def __init__(self):
        self.accession_ns = glyspace.NSGlycoinfo
        glyspace.ConjunctiveGraph.__init__(self)
        self.cache = dict()


class RDFClientTest(unittest.TestCase):
    def test_graph_reconstruction(self):
        client = LocalClient()
        client.parse(data=example_n3, format='n3')
        ref = client.get("G80903UK")
        graph = glyspace.ConjunctiveGraph().parse(
            data=ref.to_graph().serialize(format='n3'),
            format='n3')
        self.assertTrue(graph.isomorphic(client))


@skip_not_online
class GlyTouCanRDFClientTest(unittest.TestCase):
    def test_get(self):
        record = glyspace.get("G00034ND")
        self.assertEqual(record.has_primary_id, u'G00034ND')
        self.assertEqual(round(tree(record).mass()), 4104.0)
        self.assertEqual(record.exact_match.structure_, record.structure_)
        # matches = glyspace.get(glyspace.NSSKOS.exactMatch)
        # self.assertTrue(len(matches.exact_match) > 0)

    def test_from_taxon(self):
        self.assertTrue(len(glyspace.client.from_taxon(9606, limit=10)) == 10)

    def test_with_motif(self):
        self.assertTrue(len(glyspace.structures_with_motif("G00026MO", limit=10)) == 10)

    def test_structure(self):
        self.assertAlmostEqual(glyspace.structure("G00026MO").mass(), 910.327, 2)


if __name__ == '__main__':
    unittest.main()



import pkg_resources
import json


def omit_slice(seq, i):
    if i == 0:
        return seq[1:]
    return seq[:i] + seq[i + 1:]


class SynonymIndex(dict):
    def __init__(self, data):
        for group in data:
            for i, name in enumerate(group):
                self[name] = omit_slice(group, i)


class MonosaccharideSynonymIndex(SynonymIndex):
    def __init__(self, stream=None):
        if stream is None:
            data_buffer = pkg_resources.resource_string(__name__, "data/monosaccharide_synonyms.json")
            if isinstance(data_buffer, bytes):
                data_buffer = data_buffer.decode("utf-8")
        super(MonosaccharideSynonymIndex, self).__init__(json.loads(data_buffer))


#: A mapping of monosaccharide names to their synonyms
monosaccharides = MonosaccharideSynonymIndex()



import math
import operator
import itertools

from collections import deque, defaultdict

from glypy.algorithms.similarity import monosaccharide_similarity
from glypy.utils import make_struct, root, groupby, tree as treep
from glypy.structure import Glycan
from glypy.structure.monosaccharide import depth


#: Results Container for :func:`maximum_common_subgraph`
_MaximumCommonSubtreeResults = make_struct(
    "MaximumCommonSubtreeResults", ("score", "tree", "similarity_matrix"))

class MaximumCommonSubtreeResults(_MaximumCommonSubtreeResults):
    """Holds a maximum common subgraph solution

    Attributes
    ----------
    score: float
        The alignment score between the trees compared
    similarity_matrix: :class:`list` of :class:`list` of :class:`float`
        A simple dynamic programming solution matrix describing the alignment
        at each position.
    tree: :class:`~.Glycan`
        The maximum common subgraph, extracted as a separate :class:`~.Glycan`
        object.
    """
    __slots__ = ()


class MaximumCommonSubgraphSolver(object):
    '''
    Find the maximum common subgraph between :attr:`seq_a` and :attr:`seq_b`.

    Attributes
    ----------
    seq_a: :class:`~.Glycan`
    seq_b: :class:`~.Glycan`
    exact: bool
        Whether to use exact equality or fuzzy equality

    References
    ----------
    [1] K. F. Aoki, A. Yamaguchi, Y. Okuno, T. Akutsu, N. Ueda, M. Kanehisa, and H. Mamitsuka,
    "Efficient tree-matching methods for accurate carbohydrate database queries."
    Genome Inform. Jan. 2003.
    '''
    def __init__(self, seq_a, seq_b, exact=True):
        self.seq_a = seq_a
        self.seq_b = seq_b
        self.exact = exact
        self.solution_matrix = [
            [0. for i in range(len(seq_b))] for j in range(len(seq_a))]
        self.solution = None

        self.fit()

    def compare_nodes(self, node_a, node_b, visited=None):
        '''
        Score the pair of `node_a` and `node_b` and all pairings of their children.

        Returns
        -------
        float:
            The similarity score between the input nodes
        '''
        if visited is None:
            visited = set()
        score = 0.
        if (node_a.id, node_b.id) in visited:  # pragma: no cover
            return score
        visited.add((node_a.id, node_b.id))

        observed, expected = monosaccharide_similarity(
            node_a, node_b, include_substituents=True, include_children=False)
        if self.exact:
            if observed == expected:
                score += 1
        else:
            score += observed / float(expected)
        for child_a, child_b in itertools.product((ch_a for p, ch_a in node_a.children()),
                                                  (ch_b for p, ch_b in node_b.children())):
            score += self.compare_nodes(child_a, child_b, visited=visited)
        return score

    def fit(self):
        for i, a_node in enumerate(self.seq_a):
            for j, b_node in enumerate(self.seq_b):
                res = self.compare_nodes(a_node, b_node)
                self.solution_matrix[i][j] = res
        score, ix_a, ix_b = self._find_max_of_matrix(self.solution_matrix)
        node_a = self.seq_a[ix_a]
        node_b = self.seq_b[ix_b]
        self.solution = MaximumCommonSubtreeResults(
            score,
            self._extract_maximum_common_subgraph(node_a, node_b),
            self.solution_matrix)

    def _find_max_of_matrix(self, solution_matrix):
        '''
        Given the `solution_matrix`, find the coordinates of the maximum score

        Returns
        -------
        float:
            The maximum similarity score
        int:
            The index of the maximum similarity in `seq_a`
        int:
            The index of the maximum similarity in `seq_b`
        '''
        ix_a = ix_b = score = 0
        for i in range(len(solution_matrix)):
            for j in range(len(solution_matrix[0])):
                if solution_matrix[i][j] > score:
                    score = solution_matrix[i][j]
                    ix_a = i
                    ix_b = j
        return score, ix_a, ix_b

    def _coordinates_for(self, solution_matrix, value):
        solutions = []
        for i in range(len(solution_matrix)):
            for j in range(len(solution_matrix[0])):
                if solution_matrix[i][j] == value:
                    solutions.append((i, j))
        return solutions

    def _extract_maximum_common_subgraph(self, node_a, node_b):
        '''
        Given a pair of matched starting nodes from two separate glycan structures,
        traverse them together, copying the best matching branches into a new |Glycan|
        object.

        Parameters
        ----------
        node_a: :class:`~.Monosaccharide`
        node_b: :class:`~.Monosaccharide`
        exact: bool
            Whether or not to take exact matches, or just take the best
            pairing. If `exact` = |True| and there is no exact match, the
            branch will terminate.

        Returns
        -------
        :class:`~.Glycan`
        '''
        root_ = node_a.clone()
        node_stack = [(root_, node_a, node_b)]
        b_taken = set()
        index = {node_a.id: root_}
        while len(node_stack) > 0:
            mcs_node, node_a, node_b = node_stack.pop()
            for a_pos, a_child in node_a.children():
                matched_node = None
                score_pairs = {}
                if len(node_b.links) == 0:
                    continue
                for b_pos, b_child in node_b.children():
                    if b_child.id in b_taken:
                        continue
                    observed, expected = monosaccharide_similarity(
                        a_child, b_child, include_children=True)
                    if self.exact and observed == expected:
                        matched_node = b_child
                        break
                    else:
                        score_pairs[b_child.id] = (expected - observed, b_child)
                if not self.exact and len(score_pairs) > 0:
                    score, contestant = min(score_pairs.values(), key=lambda x: x[0])
                    cont_depth = depth(contestant)
                    for diff, node in score_pairs.values():
                        if diff == score:
                            node_depth = depth(node)
                            if cont_depth < node_depth:
                                contestant = node
                                cont_depth = node_depth
                    matched_node = contestant

                if matched_node is None:
                    continue

                b_taken.add(matched_node.id)
                if a_child.id in index:
                    terminal = index[a_child.id]
                else:
                    terminal = index[a_child.id] = a_child.clone()
                link = [
                    link for link in node_a.links[a_pos] if link.is_child(a_child)][0]
                link.clone(mcs_node, terminal)
                node_stack.append((terminal, a_child, matched_node))

        return Glycan(root_)

    @classmethod
    def maximum_common_subgraph(cls, seq_a, seq_b, exact=True):
        '''
        Find the maximum common subgraph between `seq_a` and `seq_b`.

        Parameters
        ----------
        seq_a: :class:`~.Glycan`
        seq_b: :class:`~.Glycan`
        exact: bool
            Whether to use exact equality or fuzzy equality

        Returns
        -------
        :class:`MaximumCommonSubtreeResults`

        References
        ----------
        [1] K. F. Aoki, A. Yamaguchi, Y. Okuno, T. Akutsu, N. Ueda, M. Kanehisa, and H. Mamitsuka,
        "Efficient tree-matching methods for accurate carbohydrate database queries."
        Genome Inform. Jan. 2003.
        '''
        inst = cls(seq_a, seq_b, exact=exact)
        return inst.solution


maximum_common_subgraph = MaximumCommonSubgraphSolver.maximum_common_subgraph


def n_saccharide_similarity(self, other, n=2, exact=False):
    """Calculate n-saccharide similarity between two structures

    Parameters
    ----------
    self: :class:`~.Glycan`
    other: :class:`~.Glycan`
    n: int
        Size of the fragment saccharide to consider. Defaults to *2*
    exact: bool
        Whether to use :meth:`Glycan.exact_ordering_equality` or :meth:`Glycan.topological_equality`.
        Defaults to `exact_ordering_equality`

    Returns
    -------
    score: float
        How similar these structures are at the n-saccharide level. Ranges between 0 and 1.0 where
        1.0 is exactly the same, while 0.0 means no shared n-saccharides.

    References
    ----------
    [1] K. F. Aoki, A. Yamaguchi, Y. Okuno, T. Akutsu, N. Ueda, M. Kanehisa, and H. Mamitsuka,
    "Efficient tree-matching methods for accurate carbohydrate database queries."
    Genome Inform. Jan. 2003.
    """
    _len = len
    _id = id

    comparator = Glycan.exact_ordering_equality
    if not exact:
        comparator = Glycan.topological_equality

    self_n_saccharides = list(subtree.tree for subtree in self.substructures(
        max_cleavages=max(self, key=operator.methodcaller('degree')).degree())
        if _len(subtree.tree) == n)

    other_n_saccharides = list(subtree.tree for subtree in other.substructures(
        max_cleavages=max(other, key=operator.methodcaller('degree')).degree())
        if _len(subtree.include_nodes) == n)

    n_sacch_max = max(len(self_n_saccharides), len(other_n_saccharides))
    matched = 0.
    paired = set()
    for stree in self_n_saccharides:
        for otree in other_n_saccharides:
            tid = _id(otree)
            if tid in paired:
                continue
            if comparator(stree, otree):
                matched += 1
                paired.add(tid)
    return matched / n_sacch_max


def distinct_fragments(self, other, fragmentation_parameters=None):
    """Compute the set of distinct masses observed between two structures

    Parameters
    ----------
    self: :class:`~.Glycan` or list
    other : :class:`~.Glycan` or list
        :class:`~.Glycan` objects whose fragments will be compared or lists of Fragment
        to compare.
    fragmentation_parameters : :class:`dict`, optional
        If `self` and `other` are |Glycan| objects, these parameters will be used
        to call :meth:`Glycan.fragments`.

    Returns
    -------
    set: The distinct fragment masses of `self`
    set: The distinct fragment masses of `other`
    """
    self_fragments = []
    other_fragments = []
    if isinstance(self, (tuple, list)):
        self_fragments = self
        other_fragments = other
    else:
        if fragmentation_parameters is None:
            fragmentation_parameters = dict(kind="BY", max_cleavages=1, average=False, charge=0)
        self_fragments = list(self.fragments(**fragmentation_parameters))
        other_fragments = list(other.fragments(**fragmentation_parameters))

    def grouper(fragment):
        return round(fragment.mass, 4)

    self_masses = set(groupby(self_fragments, grouper))
    other_masses = set(groupby(other_fragments, grouper))
    self_unique = self_masses - other_masses
    other_unique = other_masses - self_masses
    return self_unique, other_unique


class Treelet(object):
    """Represents a subgraph of a larger :class:`~.Glycan`, with a frontier
    of node ids which are children of the current subgraph.

    Attributes
    ----------
    frontier_ids : :class:`set`
        The id values of the nodes from the parent :class:`~.Glycan` which
        are children of members of :attr:`subtree`
    subtree : :class:`~.Glycan`
        The subgraph defining the treelet
    """

    def __init__(self, subtree, frontier_ids):
        self.subtree = subtree
        self.frontier_ids = set(frontier_ids)

    @classmethod
    def from_monosaccharide(cls, monosaccharide):
        monosaccharide = root(monosaccharide)
        subtree = Glycan(
            monosaccharide.clone(prop_id=True), index_method=None)
        frontier_ids = set()
        for pos, child in monosaccharide.children():
            frontier_ids.add(child.id)
        return cls(subtree, frontier_ids)

    def __len__(self):
        return len(self.subtree)

    def __root__(self):
        return root(self.subtree)

    def __tree__(self):
        return treep(self.subtree)

    def __eq__(self, other):
        try:
            return self.subtree == other.subtree and self.frontier_ids == other.frontier_ids
        except AttributeError:
            return treep(self) == treep(other)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.subtree)

    def canonicalize(self):
        self.subtree.canonicalize()

    def expand(self, reference, frontier_id):
        node = reference.get(frontier_id)
        new_node = node.clone(prop_id=True)
        tree = self.subtree.clone(index_method=None)
        for pos, link in node.parents(True):
            new_parent = tree.get(link.parent.id)
            link.clone(new_parent, new_node)
        new_frontier = set(self.frontier_ids)
        new_frontier.remove(frontier_id)
        for pos, child in node.children():
            new_frontier.add(child.id)
        return self.__class__(tree, new_frontier)

    def expand_all(self, reference):
        extent = []
        for node_id in self.frontier_ids:
            extent.append(self.expand(reference, node_id))
        return extent


class TreeletIterator(object):
    """Iterator over all distinct :math:`k`-treelets of :attr:`tree`, all
    unique sub-trees of :attr:`tree` with :attr:`k` monosaccharides.

    Attributes
    ----------
    k : int
        The number monosaccharides per treelet
    tree : :class:`~.Glycan`
        The glycan to extract reelets from
    distinct : bool
        Whether or not to filter out duplicate treelets
    """

    def __init__(self, tree, k, distinct=True):
        self.tree = tree
        self.k = k
        self.distinct = distinct
        self.node_queue = None
        self.seen = None
        self._init_node_queue()
        self.iterator = self._make_iterator()

    def _init_node_queue(self):
        self.node_queue = deque([root(self.tree)])
        self.seen = set()

    def get_next_set(self):
        node = self.node_queue.popleft()
        tree = Treelet.from_monosaccharide(node)
        for i, child in node.children():
            self.node_queue.append(child)
        return self._extend_tree(tree, 2)

    def _make_iterator(self):
        while self.node_queue:
            treelet_set = self.get_next_set()
            for treelet in treelet_set:
                treelet.canonicalize()
                if self.distinct and treelet in self.seen:
                    continue
                self.seen.add(treelet)
                yield treelet

    def __iter__(self):
        return self.iterator

    def next(self):
        return next(self.iterator)

    def __next__(self):
        return next(self.iterator)

    def _get_next_nodes(self, start):
        child_sets = []
        for pos, child_link in self.tree.get(start.id).children(links=True):
            child_sets.append(child_link)
        return child_sets

    def _extend_tree(self, tree, n):
        extents = tree.expand_all(self.tree)
        if self.k == n:
            for t in extents:
                yield t
        else:
            for subtreelet in extents:
                for t in self._extend_tree(subtreelet, n + 1):
                    yield t


def treelets(glycan, k, distinct=True):
    """Iterator over all distinct :math:`k`-treelets of :attr:`tree`, all
    unique sub-trees of :attr:`tree` with :attr:`k` monosaccharides.

    Parameters
    ----------
    glycan : :class:`~.Glycan`
        The glycan to extract reelets from
    k : int
        The number monosaccharides per treelet
    distinct : bool
        Whether or not to filter out duplicate treelets
    """
    for treelet in TreeletIterator(glycan, k, distinct=distinct):
        yield treep(treelet)


class TreeletEnrichmentTest(object):
    '''A test to calculate the probability that the frequency of each treelet
    from glycans in :attr:`cond1` is equal to the frequency of that treelet in
    :attr:`cond2` using the Fisher's Exact Test.

    Attributes
    ----------
    cond1: list of :class:`~.Glycan`
        Glycans from the first condition, the condition to be tested for enrichment
    cond2: list of :class:`~.Glycan`
        Glycans from the second condition, the background condition
    k: int
        The size of the treelet to use
    distinct: bool
        Whether or not to count redundant treelets. Defaults to |True|
    enrichment_probabilities: dict
        Holds the enrichment p values for each treelet

    References
    ----------
    Pevzner, P., & Shamir, R. (2011). Bioinformatics for Biologists.
    New York, NY, USA: Cambridge University Press.
    '''
    def __init__(self, cond1, cond2, k, distinct=True):
        self.cond1 = list(cond1)
        self.cond2 = list(cond2)
        self.k = k
        self.distinct = distinct
        self.total = len(self.cond1) + len(self.cond2)
        self.cond1_treelets = defaultdict(int)
        self.cond2_treelets = defaultdict(int)
        self.cond1_treelet_to_glycan = defaultdict(set)
        self.cond2_treelet_to_glycan = defaultdict(set)
        self.enrichment_probabilities = dict()
        self.fit()

    def count_treelets(self):
        for g1 in self.cond1:
            for treelet in treelets(g1, self.k, distinct=self.distinct):
                self.cond1_treelets[treelet] += 1
                self.cond1_treelet_to_glycan[treelet].add(g1)
        for g2 in self.cond2:
            for treelet in treelets(g2, self.k, distinct=self.distinct):
                self.cond2_treelets[treelet] += 1
                self.cond2_treelet_to_glycan[treelet].add(g2)

    def fisher_exact_test(self, treelet):
        M = self.total
        N = len(self.cond1)
        n_pos = len(self.cond1_treelet_to_glycan[treelet])
        n_neg = len(self.cond2_treelet_to_glycan[treelet])

        fac = math.factorial
        numerator = fac(M) / float(fac(n_pos) * fac(n_neg) * fac(N - n_pos) * fac(M - N - fac(n_neg)))
        d1 = fac(M) / float(fac(n_pos + n_neg) * fac(M - n_pos - n_neg))
        d2 = fac(M) / float(fac(M - N) * fac(N) * N)
        p = numerator / (d1 * d2)
        return p

    def fit(self):
        self.count_treelets()
        for treelet in self.cond1_treelet_to_glycan:
            self.enrichment_probabilities[treelet] = self.fisher_exact_test(treelet)

    @classmethod
    def treelet_enrichment(cls, cond1, cond2, k, distinct=True):
        '''Perform a test for each treelet to determine if it is enriched in *cond1*
        compared to *cond2*.

        Parameters
        ----------
        cond1: list of :class:`~.Glycan`
            Glycans from the first condition, the condition to be tested for enrichment
        cond2: list of :class:`~.Glycan`
            Glycans from the second condition, the background condition
        k: int
            The size of the treelet to use
        distinct: bool
            Whether or not to count redundant treelets. Defaults to |True|

        Returns
        -------
        dict
            A mapping from treelet to p value from Fisher's Exact Test for
            that treelet being found with its observed frequency in *cond1*
            if it is as common in *cond1* as in *cond2*.

        References
        ----------
        Pevzner, P., & Shamir, R. (2011). Bioinformatics for Biologists.
        New York, NY, USA: Cambridge University Press.
        '''
        inst = cls(cond1, cond2, k, distinct=distinct)
        return inst.enrichment_probabilities


treelet_enrichment = TreeletEnrichmentTest.treelet_enrichment



import unittest

import glypy
from glypy.io import glycoct
from glypy.tests.common import load, glycan, structures as raw_structures


example_multiplicity = '''
RES
1r:r1
REP
REP1:3o(3+1)2d=1-4
RES
2b:a-dman-HEX-1:5
3b:a-lman-HEX-1:5|6:d
4b:b-lxyl-PEN-1:5
LIN
1:2o(4+1)3d
2:3o(2+1)4d
'''

osvaldo_repeat_1 = '''
RES
1r:r1
REP
REP1:5o(4+1)2d=-1--1
RES
2b:b-dglc-HEX-1:5
3b:b-dgal-HEX-1:5
4b:b-dglc-HEX-1:5
5b:a-dglc-HEX-1:5
6b:a-dglc-HEX-1:5
7s:n-acetyl
8b:b-dgal-HEX-1:5
9b:a-dgro-dgal-NON-2:6|1:a|2:keto|3:d
10s:n-acetyl
LIN
1:2o(4+1)3d
2:3o(3+1)4d
3:3o(4+1)5d
4:5o(6+1)6d
5:6d(2+1)7n
6:6o(4+1)8d
7:8o(3+2)9d
8:9d(5+1)10n
'''

osvaldo_repeat_2 = '''
RES
1r:r1
REP
REP1:6o(3+1)2n=-1--1
RES
2r:r2
3r:r3
4r:r4
5r:r5
6b:b-dgal-HEX-1:5
LIN
1:2n(4+1)3n
2:3n(3+1)4n
3:4n(3+1)5n
4:5n(3+1)6d
REP2:7o(4+1)7d=-1--1
RES
7b:b-dxyl-PEN-1:5
REP3:8o(3+1)8d=-1--1
RES
8b:b-dgal-HEX-1:5
9b:b-dgal-HEX-1:5
LIN
5:8o(6+1)9d
REP4:10o(3+1)10d=-1--1
RES
10b:b-dgal-HEX-1:5
11b:b-dgal-HEX-1:5
12b:b-dgal-HEX-1:5
13b:b-dgal-HEX-1:5
LIN
6:10o(6+1)11d
7:11o(3+1)12d
8:12o(3+1)13d
REP5:14o(3+1)14d=-1--1
RES
14b:b-dgal-HEX-1:5
15b:b-dgal-HEX-1:5
16b:a-lara-PEN-1:4
17b:a-lara-PEN-1:4
LIN
9:14o(6+1)15d
10:15o(6+1)16d
11:16o(3+1)17d
'''


test_multiple_repeating_buffer = '''
RES
1b:b-dglc-HEX-1:5
2s:n-acetyl
3b:b-dglc-HEX-1:5
4s:n-acetyl
5b:b-dman-HEX-1:5
6b:a-dman-HEX-1:5
7b:b-dglc-HEX-1:5
8s:n-acetyl
9r:r1
10b:b-dgal-HEX-1:5
11b:a-dman-HEX-1:5
12b:b-dglc-HEX-1:5
13s:n-acetyl
14r:r2
15b:b-dgal-HEX-1:5
16b:a-lgal-HEX-1:5|6:d
LIN
1:1d(2+1)2n
2:1o(4+1)3d
3:3d(2+1)4n
4:3o(4+1)5d
5:5o(3+1)6d
6:6o(2+1)7d
7:7d(2+1)8n
8:7o(4+1)9n
9:9n(4+1)10d
10:5o(6+1)11d
11:11o(2+1)12d
12:12d(2+1)13n
13:12o(4+1)14n
14:14n(4+1)15d
15:1o(6+1)16d
REP
REP1:18o(4+1)17d=-1--1
RES
17b:b-dgal-HEX-1:5
18b:b-dglc-HEX-1:5
19s:n-acetyl
LIN
16:17o(3+1)18d
17:18d(2+1)19n
REP2:21o(4+1)20d=-1--1
RES
20b:b-dgal-HEX-1:5
21b:b-dglc-HEX-1:5
22s:n-acetyl
LIN
18:20o(3+1)21d
19:21d(2+1)22n


RES
1b:x-dglc-HEX-x:x
2s:n-acetyl
3b:b-dglc-HEX-1:5
4s:n-acetyl
5b:b-dman-HEX-1:5
6b:a-dman-HEX-1:5
7b:b-dglc-HEX-1:5
8s:n-acetyl
9b:b-dgal-HEX-1:5
10b:b-dglc-HEX-1:5
11s:n-acetyl
12b:b-dgal-HEX-1:5
13b:a-dman-HEX-1:5
14b:b-dglc-HEX-1:5
15s:n-acetyl
16b:b-dgal-HEX-1:5
17b:b-dglc-HEX-1:5
18s:n-acetyl
19b:b-dgal-HEX-1:5
20b:a-lgal-HEX-1:5|6:d
LIN
1:1d(2+1)2n
2:1o(4+1)3d
3:3d(2+1)4n
4:3o(4+1)5d
5:5o(3+1)6d
6:6o(2+1)7d
7:7d(2+1)8n
8:7o(4+1)9d
9:6o(4+1)10d
10:10d(2+1)11n
11:10o(4+1)12d
12:5o(6+1)13d
13:13o(2+1)14d
14:14d(2+1)15n
15:14o(4+1)16d
16:13o(6+1)17d
17:17d(2+1)18n
18:17o(4+1)19d
19:1o(6+1)20d
'''


und_case1 = '''
RES
1b:a-dman-HEX-1:5
2r:r1
3b:a-dgal-HEX-1:5
LIN
1:1o(2+1)2n
2:2n(2+1)3d
REP
REP1:4o(2+1)4d=-1--1
RES
4b:a-dglc-HEX-1:5
UND
UND1:100.0:100.0
ParentIDs:3|1
SubtreeLinkageID1:o(2+4)d
RES
5b:b-dgro-dgal-NON-0:0|1:a|2:keto|3:d
6s:n-acetyl
LIN
3:5d(5+2)6n
'''


und_case2 = '''
RES
1b:a-dman-HEX-1:5
2r:r1
3b:a-dgal-HEX-1:5
LIN
1:1o(2+1)2n
2:2n(2+1)3d
REP
REP1:4o(2+1)4d=1-5
RES
4b:a-dglc-HEX1:5
UND
UND1:60.0:80.0
ParentIDs:4
SubtreeLinkageID1:o(4+1)n
RES
5s:sulfate

'''


class GlycoCTParserTests(unittest.TestCase):
    _file_path = "./test_data/glycoct.txt"

    def test_parse_file(self):
        with open(self._file_path) as stream:
            for g in glycoct.read(stream):
                self.assertTrue(isinstance(g, glycan.Glycan))

    def test_parse_cyclical(self):
        structure = load("cyclical_glycan")
        self.assertAlmostEqual(structure.mass(), 810.2641170925)

    def test_parse_multiple_structures_in_buffer(self):
        structures = glycoct.loads(test_multiple_repeating_buffer)
        self.assertEqual(len(structures), 2)

        self.assertAlmostEqual(structures[0].mass(), structures[1].mass(), 3)
        self.assertNotEqual(*structures)

    def test_parse_repeating(self):
        structure = load("repeating_glycan")
        self.assertAlmostEqual(structure.mass(), 666.22185, 3)

        structure = load("big_repeating_glycan")
        self.assertAlmostEqual(structure.mass(), 5248.87193, 3)

        structure = glycoct.loads(example_multiplicity)
        self.assertEqual(len(structure), 4 * 3)

        structure = glycoct.loads(osvaldo_repeat_1)
        self.assertAlmostEqual(structure.mass(), 1322.449, 3)
        self.assertEqual(len(structure), 7)

        structure = glycoct.loads(osvaldo_repeat_2)
        self.assertAlmostEqual(structure.mass(), 1872.612751, 3)
        self.assertEqual(len(structure), 12)

    def test_parse_undetermined(self):
        structure = glycoct.loads(und_case1)
        for link in structure.link_index:
            if link.is_ambiguous():
                self.assertTrue(len(link.child.stem) > 1)

        structure = glycoct.loads(und_case2)
        sulfate_count = 0
        for node in structure:
            for i, subst in node.substituents():
                if subst.name == 'sulfate':
                    sulfate_count += 1
        self.assertEqual(sulfate_count, 5)

    def test_composition(self):
        gc = glypy.GlycanComposition.parse("{Hex:5; Hex2NAc:4; Neu5Ac:2; @sulfate: 1}")
        self.assertEqual(gc, glycoct.loads(glycoct.dumps(gc)))

    def test_ordered_serialization(self):
        for acc in ['G58143RL', 'G82388RB', 'G28839WC', 'G37369XO', 'G36221RT', 'G27293OK',
                    'G62831KM', 'G65832JS', 'G07337US', 'G01120GS']:
            text = raw_structures[acc].strip()
            structure = glycoct.loads(text)
            retext = glycoct.dumps(structure).strip()
            assert text == retext, "Failed to match %s" % acc


if __name__ == '__main__':
    unittest.main()



import unittest

import glypy
from glypy.composition import composition_transform
from glypy.structure import glycan_composition
from glypy import monosaccharides, Substituent, glycans

from .common import pickle

water_mass = glypy.Composition("H2O").mass

GlycanComposition = glycan_composition.GlycanComposition
MonosaccharideResidue = glycan_composition.MonosaccharideResidue

FrozenGlycanComposition = glycan_composition.FrozenGlycanComposition
FrozenMonosaccharideResidue = glycan_composition.FrozenMonosaccharideResidue


class IUPACLiteTests(unittest.TestCase):
    def test_parse_identity(self):
        items = ["Glc2NAc", "Neu5Ac", "Fuc", "6-d-Hex", "#Ammonia#H3N1"]
        for item in items:
            term = glycan_composition.from_iupac_lite(item)
            deparse = glycan_composition.to_iupac_lite(term)
            self.assertEqual(item, deparse)


class MonosaccharideResidueTests(unittest.TestCase):
    def test_from_monosaccharide(self):
        for base in ["GlcNAc", "NeuAc", "Fuc"]:
            base = monosaccharides[base]
            residue = glycan_composition.MonosaccharideResidue.from_monosaccharide(base)
            self.assertAlmostEqual(base.mass() - water_mass, residue.mass(), 3)

    def test_clone_coherence(self):
        for base in ["GlcNAc", "NeuAc", "Fuc"]:
            base = monosaccharides[base]
            residue = glycan_composition.MonosaccharideResidue.from_monosaccharide(base)
            dup = residue.clone()
            self.assertAlmostEqual(dup.mass(), residue.mass(), 3)

    def test_derivatize(self):
        reference = {"GlcNAc": 245.1263, "Fuc": 174.0892, "NeuAc": 361.1737}
        for name in ["GlcNAc", "NeuAc", "Fuc"]:
            base = monosaccharides[name]
            residue = glycan_composition.MonosaccharideResidue.from_monosaccharide(base)
            composition_transform.derivatize(base, "methyl")
            composition_transform.derivatize(residue, "methyl")
            self.assertAlmostEqual(
                base.mass() - (water_mass + 2 * (Substituent("methyl").mass() - glypy.Composition("H").mass * 2)),
                residue.mass(), 3)
            self.assertAlmostEqual(reference[name], residue.mass(), 3)

            base = monosaccharides[name]
            composition_transform.derivatize(base, "methyl")
            residue = glycan_composition.MonosaccharideResidue.from_monosaccharide(base)

            self.assertAlmostEqual(
                base.mass() - (water_mass + 2 * (Substituent("methyl").mass() - glypy.Composition("H").mass * 2)),
                residue.mass(), 3)

            self.assertAlmostEqual(reference[name], residue.mass(), 3)

    def test_acidic_special_case(self):
        acidic_hexose1 = glycan_composition.MonosaccharideResidue.from_iupac_lite("HexA")
        acidic_hexose2 = glycan_composition.MonosaccharideResidue.from_iupac_lite("aHex")
        hexose = glycan_composition.MonosaccharideResidue.from_iupac_lite("Hex")
        self.assertAlmostEqual(acidic_hexose1.mass() - hexose.mass(), 13.9792, 3)
        self.assertAlmostEqual(acidic_hexose2.mass() - hexose.mass(), 13.9792, 3)

    def test_pickle(self):
        case = glycan_composition.MonosaccharideResidue.from_iupac_lite("HexNAc")
        test = pickle.loads(pickle.dumps(case))
        self.assertEqual(case, test)


class FrozenMonosaccharideResidueTests(unittest.TestCase):
    def test_from_monosaccharide(self):
        for base in ["GlcNAc", "NeuAc", "Fuc"]:
            base = monosaccharides[base]
            residue = glycan_composition.FrozenMonosaccharideResidue.from_monosaccharide(base)
            self.assertAlmostEqual(base.mass() - water_mass, residue.mass(), 3)

    def test_clone_coherence(self):
        for base in ["GlcNAc", "NeuAc", "Fuc"]:
            base = monosaccharides[base]
            residue = glycan_composition.FrozenMonosaccharideResidue.from_monosaccharide(base)
            dup = residue.clone()
            self.assertAlmostEqual(dup.mass(), residue.mass(), 3)

    def test_derivatize(self):
        reference = {"GlcNAc": 245.1263, "Fuc": 174.0892, "NeuAc": 361.1737}
        with self.assertRaises(glycan_composition.FrozenError):
            for name in ["GlcNAc", "NeuAc", "Fuc"]:
                base = monosaccharides[name]
                residue = glycan_composition.FrozenMonosaccharideResidue.from_monosaccharide(base)
                composition_transform.derivatize(base, "methyl")
                composition_transform.derivatize(residue, "methyl")
                self.assertAlmostEqual(
                    base.mass() - (water_mass + 2 * (Substituent("methyl").mass() - glypy.Composition("H").mass * 2)),
                    residue.mass(), 3)
                self.assertAlmostEqual(reference[name], residue.mass(), 3)

                base = monosaccharides[name]
                composition_transform.derivatize(base, "methyl")
                residue = glycan_composition.FrozenMonosaccharideResidue.from_monosaccharide(base)

                self.assertAlmostEqual(
                    base.mass() - (water_mass + 2 * (Substituent("methyl").mass() - glypy.Composition("H").mass * 2)),
                    residue.mass(), 3)

                self.assertAlmostEqual(reference[name], residue.mass(), 3)

    def test_acidic_special_case(self):
        acidic_hexose1 = glycan_composition.FrozenMonosaccharideResidue.from_iupac_lite("HexA")
        acidic_hexose2 = glycan_composition.FrozenMonosaccharideResidue.from_iupac_lite("aHex")
        hexose = glycan_composition.FrozenMonosaccharideResidue.from_iupac_lite("Hex")
        self.assertAlmostEqual(acidic_hexose1.mass() - hexose.mass(), 13.9792, 3)
        self.assertAlmostEqual(acidic_hexose2.mass() - hexose.mass(), 13.9792, 3)

    def test_pickle(self):
        case = glycan_composition.MonosaccharideResidue.from_iupac_lite("HexNAc")
        test = pickle.loads(pickle.dumps(case))
        self.assertEqual(case, test)


class GlycanCompositionTests(unittest.TestCase):
    GlycanCompositionType = GlycanComposition

    def test_from_glycan(self):
        glyc = glycans["N-Linked Core"]
        comp = self.GlycanCompositionType.from_glycan(glyc)
        self.assertAlmostEqual(glyc.mass(), comp.mass(), 3)

    def test_derivatize(self):
        glyc = glycans["N-Linked Core"]
        composition_transform.derivatize(glyc, "methyl")
        comp = self.GlycanCompositionType.from_glycan(glyc)
        self.assertAlmostEqual(glyc.mass(), comp.mass(), 3)

        comp = self.GlycanCompositionType.from_glycan(glycans["N-Linked Core"])
        composition_transform.derivatize(comp, "methyl")
        self.assertAlmostEqual(glyc.mass(), comp.mass(), 3)
        self.assertAlmostEqual(self.GlycanCompositionType.parse(comp).mass(), comp.mass(), 3)

    def test_contains(self):
        glyc = glycans["N-Linked Core"]
        comp = self.GlycanCompositionType.from_glycan(glyc)
        self.assertTrue("Man" in comp)

    def test_serialize(self):
        ref = '{Man:3; Glc2NAc:2}'
        glyc = glycans["N-Linked Core"]
        comp = self.GlycanCompositionType.from_glycan(glyc)
        self.assertEqual(comp.serialize(), ref)

        ref2 = '{Hex:3; Hex2NAc:2}'
        comp.drop_stems()
        self.assertEqual(comp.serialize(), ref2)

    def test_parse(self):
        ref = '{Man:3; Glc2NAc:2}'
        glyc = glycans["N-Linked Core"]
        comp = self.GlycanCompositionType.from_glycan(glyc)

        comp2 = self.GlycanCompositionType.parse(ref)
        self.assertEqual(comp, comp2)

    def test_update(self):
        ref = '{Man:3; Glc2NAc:2}'
        comp = self.GlycanCompositionType.parse(ref)
        self.assertEqual(comp["Man"], 3)
        comp.update({"Man": 1})
        self.assertEqual(comp["Man"], 1)
        comp.update(Man=2)
        self.assertEqual(comp["Man"], 2)

    def test_arithmetic(self):
        ref = '{Man:3; Glc2NAc:2}'
        comp = self.GlycanCompositionType.parse(ref)

        comp["Man"] += 5
        self.assertEqual(comp["Man"], 8)
        comp2 = self.GlycanCompositionType(Neu5NAc=2, Fuc=1, Man=1)
        self.assertEqual(comp2 + comp, self.GlycanCompositionType(
            Neu5NAc=2, Fuc=1, Man=9, Glc2NAc=2))
        comp -= {'Man': 5}
        self.assertEqual(comp, self.GlycanCompositionType(
            Man=4, Glc2NAc=2) - self.GlycanCompositionType(Man=1, Glc2NAc=0))
        comp *= 2
        self.assertEqual(comp, (self.GlycanCompositionType(
            Man=4, Glc2NAc=2) - self.GlycanCompositionType(Man=1, Glc2NAc=0)) * 2)

        comp3 = comp2.clone()
        comp3 += comp2
        self.assertEqual(comp3, comp2 * 2)

    def test_total_composition(self):
        ref = '{Man:3; Glc2NAc:2}'
        comp = self.GlycanCompositionType.parse(ref)
        glyc = glycans["N-Linked Core"]
        self.assertEqual(comp.total_composition(), glyc.total_composition())

    def test_drops(self):
        ref = '{Man:3; Glc2NAc:2}'
        comp = self.GlycanCompositionType.parse(ref)
        self.assertEqual(comp["Man"], 3)
        self.assertEqual(comp["Glc2NAc"], 2)
        comp.drop_positions()
        self.assertEqual(comp["GlcNAc"], 2)
        self.assertEqual(comp["Glc2NAc"], 0)

        comp.drop_stems()
        self.assertEqual(comp["HexNAc"], 2)
        self.assertEqual(comp["GlcNAc"], 0)
        self.assertEqual(comp["Glc2NAc"], 0)
        self.assertEqual(comp["Hex"], 3)
        self.assertEqual(comp["Man"], 0)

    def test_parse_derivatized_reduced(self):
        x = self.GlycanCompositionType.parse('{Fuc^Me:1; Hex^Me:5; HexNAc^Me:4; Neu5NAc^Me:1}$C1H4')
        self.assertAlmostEqual(x.mass(), 2598.3402, 4)

    def test_pickle(self):
        x = self.GlycanCompositionType.parse('{Fuc:1; Hex:5; HexNAc:4; Neu5NAc:1; @sulfate:1}')
        self.assertEqual(x, pickle.loads(pickle.dumps(x)))

    def test_unsaturated(self):
        x = self.GlycanCompositionType.parse("{Xyl:1; a,enHex:1; Hex:1; HexS:1; aHex:1; HexNAc(S):1}")
        self.assertAlmostEqual(x.mass(), 1171.205, 3)

    def test_query(self):
        x = self.GlycanCompositionType.parse("{GlcNAc:5; Man:3; Gal:2; NeuAc:2}")
        self.assertEqual(x.query("Man"), 3)
        self.assertEqual(x.query("Hex"), 5)

    def test_reinterpret(self):
        x = self.GlycanCompositionType.parse("{GlcNAc:5; Man:3; Gal:2; NeuAc:2}")
        y = x.reinterpret([
            x._key_parser("Hex"),
            x._key_parser("dHex")
        ])
        self.assertAlmostEqual(x.mass(), y.mass(), 5)
        self.assertEqual(len(y), 3)

    def test_equality(self):
        x = self.GlycanCompositionType.parse("{Hex:9; HexNAc:2}")
        self.assertEqual(x, str(x))
        self.assertEqual(x, x)
        y = self.GlycanCompositionType.parse("{Hex:9; HexNAc:3}")
        self.assertNotEqual(x, y)
        self.assertNotEqual(x, str(y))


class FrozenGlycanCompositionTests(GlycanCompositionTests):
    GlycanCompositionType = glycan_composition.FrozenGlycanComposition

    def test_derivatize(self):
        glyc = glycans["N-Linked Core"]
        with self.assertRaises(glycan_composition.FrozenError):
            composition_transform.derivatize(glyc, "methyl")
            comp = self.GlycanCompositionType.from_glycan(glyc)
            self.assertAlmostEqual(glyc.mass(), comp.mass(), 3)

            comp = self.GlycanCompositionType.from_glycan(glycans["N-Linked Core"])
            composition_transform.derivatize(comp, "methyl")
            self.assertAlmostEqual(glyc.mass(), comp.mass(), 3)
            self.assertAlmostEqual(self.GlycanCompositionType.parse(comp).mass(), comp.mass(), 3)

    def test_parse_derivatized(self):
        glyc = glycans["N-Linked Core"]
        composition_transform.derivatize(glyc, "methyl")
        comp = GlycanComposition.from_glycan(glycans["N-Linked Core"])
        composition_transform.derivatize(comp, "methyl")
        self.assertAlmostEqual(glyc.mass(), comp.mass(), 3)
        frozen_comp = self.GlycanCompositionType.parse(comp)
        self.assertAlmostEqual(glyc.mass(), frozen_comp.mass(), 3)

    def test_serialize(self):
        ref = '{Man:3; Glc2NAc:2}'
        glyc = glycans["N-Linked Core"]
        comp = self.GlycanCompositionType.from_glycan(glyc)
        self.assertEqual(comp.serialize(), ref)

    def test_drops(self):
        ref = '{Man:3; Glc2NAc:2}'
        comp = self.GlycanCompositionType.parse(ref)
        self.assertEqual(comp["Man"], 3)
        self.assertEqual(comp["Glc2NAc"], 2)
        with self.assertRaises(glycan_composition.FrozenError):
            comp.drop_positions()


class SubstituentResidueTests(unittest.TestCase):
    def test_parse(self):
        residue = glycan_composition.from_iupac_lite("@n_acetyl")
        hexose = MonosaccharideResidue.from_iupac_lite("Hex")
        hexnac = MonosaccharideResidue.from_iupac_lite("HexNAc")
        self.assertAlmostEqual(residue.mass(), hexnac.mass() - hexose.mass(), 3)

    def test_to_iupac_lite(self):
        reference = "@n_acetyl"
        residue = glycan_composition.from_iupac_lite(reference)
        self.assertEqual(reference, str(residue))

    def test_hashable(self):
        c = dict()
        residue = glycan_composition.from_iupac_lite("@n_acetyl")
        c[residue] = 5
        self.assertEqual(c[glycan_composition.from_iupac_lite("@n_acetyl")], 5)

    def test_equality(self):
        residue = glycan_composition.from_iupac_lite("@n_acetyl")
        self.assertEqual(residue, glycan_composition.from_iupac_lite("@n_acetyl"))
        self.assertNotEqual(residue, "n_acetyl")


if __name__ == '__main__':
    unittest.main()
